# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Collection of Zarr groups
=========================
"""
from __future__ import annotations

from typing import Any, Iterable, Iterator, Literal, NoReturn, Sequence
import datetime
import functools
import io
import json
import logging
import os
import types
import warnings

import dask.bag.core
import dask.distributed
import dask.utils
import fsspec
import xarray

from .. import (
    dask_utils,
    dataset,
    fs_utils,
    merging,
    meta,
    partitioning,
    storage,
    sync,
)
from .abc import PartitionFilter, ReadOnlyCollection
from .callable_objects import UpdateCallable, WrappedPartitionCallable
from .detail import (
    _insert,
    _try_infer_callable,
    _wrap_update_func,
    _wrap_update_func_with_overlap,
)

#: Module logger.
_LOGGER: logging.Logger = logging.getLogger(__name__)


def _write_immutable_dataset(
    zds: dataset.Dataset,
    axis: str,
    path: str,
    fs: fsspec.AbstractFileSystem,
    synchronizer: sync.Sync,
) -> None:
    """Write the immutable dataset.

    Args:
        zds: The dataset to write.
        axis: The partitioning axis.
        path: The path to the immutable dataset.
        fs: The file system that the partition is stored on.
    """
    immutable_dataset = zds.select_variables_by_dims((axis, ), predicate=False)
    assert len(immutable_dataset.variables) != 0, (
        'The dataset to insert does not contain any variable '
        'that is not split.')
    _LOGGER.info('Creating the immutable dataset: %s', path)
    storage.write_zarr_group(immutable_dataset,
                             path,
                             fs,
                             synchronizer,
                             distributed=False)


def _infer_callable(
    collection: Collection,
    func: UpdateCallable,
    filters: PartitionFilter | None,
    delayed,
    selected_variables: Iterable[str] | None,
    *args,
    **kwargs,
) -> tuple[str, ...]:
    try:
        one_partition: str = next(collection.partitions(filters=filters))
    except StopIteration:
        return tuple()

    with collection.synchronizer:
        zds: dataset.Dataset = storage.open_zarr_group(
            one_partition,
            collection.fs,
            delayed=delayed,
            selected_variables=selected_variables)
    func_result: dict[str, Any]
    func_result = _try_infer_callable(func, zds,
                                      collection.partition_properties.dim,
                                      *args, **kwargs)
    unknown_variables: set[str] = set(func_result) - set(
        collection.metadata.variables.keys())
    if len(unknown_variables):
        raise ValueError(f'Unknown variables: {unknown_variables}')
    return tuple(func_result)


class Collection(ReadOnlyCollection):
    """This class manages a collection of files in Zarr format stored in a set
    of subdirectories. These subdirectories split the data, by cycles or dates
    for example, in order to optimize access and updates, deletion or addition
    of new data.

    Args:
        axis: The axis of the collection. This is the dimension along which the
            data is partitioned.
        ds: The dataset containing the collection. This dataset is used to
            create the metadata of the collection, which is used to validate
            datasets that are inserted in the collection.
        partition_handler: The partitioning strategy for the collection. This
            is an instance of a subclass of
            :py:class:`zarr_partitioning.PartitionHandler`.
        partition_base_dir: The base directory for the collection. This is the
            directory where the subdirectories containing the partitioned data
            are stored.
        mode: The mode of the collection. This can be either 'r' (read-only) or
            'w' (write). In read-only mode, the collection can only be read and
            no data can be inserted or modified. In write mode, the collection
            can be read and modified.
        filesystem: The filesystem to use for the collection. This is an
            instance of a subclass of :py:class:`fsspec.AbstractFileSystem`.
        synchronizer: The synchronizer to use for the collection. This is an
            instance of a subclass of
            :py:class:`zarr_synchronizer.Synchronizer`.

    Raises:
        ValueError:
            If the axis does not exist in the dataset, if the partition key is
            not defined in the dataset or if the access mode is not supported.

    Notes:
        Normally, this class is not instantiated directly but through the
        :py:meth:`create_collection <zcollection.create_collection>` and
        :py:meth:`open_collection <zcollection.open_collection>` methods of this
        library.
    """

    def __init__(
        self,
        axis: str,
        ds: meta.Dataset,
        partition_handler: partitioning.Partitioning,
        partition_base_dir: str,
        *,
        mode: Literal['r', 'w'] | None = None,
        filesystem: fsspec.AbstractFileSystem | str | None = None,
        synchronizer: sync.Sync | None = None,
    ) -> None:
        super().__init__(axis=axis,
                         ds=ds,
                         partition_handler=partition_handler,
                         partition_base_dir=partition_base_dir,
                         mode=mode,
                         filesystem=filesystem,
                         synchronizer=synchronizer)

        if self.mode == 'r':
            # pylint: disable=method-hidden
            # These methods are overloaded when the collection is opened in
            # readonly.
            self._read_only_mode()
            # pylint: enable=method-hidden
        else:
            self._write_config(skip_if_exists=True)

    def __str__(self) -> str:
        return (f'<{self.__class__.__name__} '
                f'filesystem={self.fs.__class__.__name__!r}, '
                f'partition_base_dir={self.partition_properties.dir!r}'
                f'mode={self.mode!r}>')

    @staticmethod
    def _unsupported_operation(*args, **kwargs) -> NoReturn:
        """Raise an exception if the operation is not supported."""
        raise io.UnsupportedOperation('not writable')

    def _read_only_mode(self) -> None:
        """Set the unsupported methods to raise an exception when the
        collection is opened in read-only mode."""
        # Set each unsupported method to raise an exception.
        for item in [
                'add_variable',
                'drop_partitions',
                'drop_variable',
                'insert',
                'update',
        ]:
            assert hasattr(self, item), f'{item} is not a known method.'
            setattr(self, item,
                    types.MethodType(Collection._unsupported_operation, self))

    def _write_config(self, skip_if_exists: bool = False) -> None:
        """Write the configuration file."""
        base_dir: str = self.partition_properties.dir
        config: str = self._config(base_dir)
        exists: bool = self.fs.exists(config)

        if exists and skip_if_exists:
            return

        _LOGGER.info(
            "Updating collection's configuration: %s"
            if exists else 'Creating the collection: %s', config)

        self.fs.makedirs(base_dir, exist_ok=True)

        params = {
            'axis': self.axis,
            'dataset': self.metadata.get_config(),
            'partitioning': self.partitioning.get_config(),
        }

        with self.fs.open(config, mode='w') as stream:
            json.dump(params, stream, indent=4)  # type: ignore[arg-type]

    def is_readonly(self) -> bool:
        """Return True if the collection is read-only."""
        return self.mode == 'r'

    @classmethod
    def from_config(
        cls,
        path: str,
        *,
        mode: Literal['r', 'w'] | None = None,
        filesystem: fsspec.AbstractFileSystem | str | None = None,
        synchronizer: sync.Sync | None = None,
    ) -> Collection:
        """Open a Collection described by a configuration file.

        Args:
            path: The path to the configuration file.
            mode: The mode of the collection. This can be either 'r'
                (read-only) or 'w' (write).
            filesystem: The filesystem to use for the collection. This is an
                instance of a subclass of
                :py:class:`fsspec.AbstractFileSystem`.
            synchronizer: The synchronizer to use for the collection. This is
                an instance of a subclass of
                :py:class:`zarr_synchronizer.Synchronizer`.

        Returns:
            The collection.

        Raises:
            ValueError:
                If the provided directory does not contain a valid collection
                configuration file.
        """
        _LOGGER.info('Opening collection: %r', path)
        fs: fsspec.AbstractFileSystem = fs_utils.get_fs(filesystem)
        config: str = cls._config(path)
        if not fs.exists(config):
            raise ValueError(f'zarr collection not found at path {path!r}')
        with fs.open(config) as stream:
            data: dict[str, Any] = json.load(stream)
        return Collection(
            data['axis'],
            meta.Dataset.from_config(data['dataset']),
            partitioning.get_codecs(data['partitioning']),
            path,
            mode=mode or 'r',
            filesystem=fs,
            synchronizer=synchronizer,
        )

    # pylint: disable=method-hidden
    def insert(
        self,
        ds: xarray.Dataset | dataset.Dataset,
        *,
        merge_callable: merging.MergeCallable | None = None,
        npartitions: int | None = None,
        validate: bool = False,
    ) -> None:
        """Insert a dataset into the collection.

        Args:
            ds: The dataset to insert. It can be either an xarray.Dataset or a
                dataset.Dataset object.
            merge_callable: A function to use to merge the existing data set
                already stored in partitions with the new partitioned data. If
                None, the new partitioned data overwrites the existing
                partitioned data.
            npartitions: The maximum number of partitions to process in
                parallel. By default, partitions are processed one by one.

                .. note::

                    When inserting partitions, Dask parallelizes the writing of
                    each partition across its workers. Additionally, the writing
                    of variables within a partition is parallelized on the
                    worker responsible for inserting that partition, using
                    multiple threads. If you're using a single Dask worker,
                    partition insertion will happen sequentially and changing
                    this parameter will have no effect.

            validate: Whether to validate dataset metadata before insertion
                or not.

        Raises:
            ValueError:
                If the dataset does not match the definition of the collection.

        Warns:
            UserWarning:
                If two different partitions use the same file (chunk), the
                library that handles the storage of chunked arrays (HDF5,
                NetCDF, Zarr, etc.) must be compatible with concurrent access.

        Notes:
            Each worker will process a set of independent partitions. However,
            be careful, two different partitions can use the same file (chunk),
            therefore, the library that handles the storage of chunked arrays
            (HDF5, NetCDF, Zarr, etc.) must be compatible with concurrent
            access.
        """
        # pylint: disable=method-hidden
        if isinstance(ds, xarray.Dataset):
            ds = dataset.Dataset.from_xarray(ds)

        _LOGGER.info('Inserting of a %s dataset in the collection',
                     dask.utils.format_bytes(ds.nbytes))

        missing_variables: tuple[str, ...] = self.metadata.missing_variables(
            ds.metadata())
        for item in missing_variables:
            variable: meta.Variable = self.metadata.variables[item]
            ds.add_variable(variable)

        if validate and ds.metadata() != self.metadata:
            raise ValueError(
                "Provided dataset's metadata do not match the collection's ones"
            )
        ds = ds.set_for_insertion(self.metadata)
        client: dask.distributed.Client = dask_utils.get_client()

        # If the dataset contains variables that should not be partitioned.
        if self._immutable is not None:

            # On the first call, we store the immutable variables in
            # a directory located at the root of the collection.
            if not self.fs.exists(self._immutable):
                _write_immutable_dataset(ds, self.axis, self._immutable,
                                         self.fs, self.synchronizer)

            # Remove the variables that should not be partitioned.
            ds = ds.select_variables_by_dims((self.axis, ))

        # Process the partitions to insert or update by batches to avoid
        # memory issues.
        partitions = tuple(
            self.partitioning.split_dataset(ds, self.partition_properties.dim))

        if npartitions is not None:
            if npartitions < 1:
                raise ValueError('The number of partitions must be positive')
            npartitions = len(partitions) // npartitions + 1

        scattered_ds: Any = client.scatter(ds)
        for sequence in dask_utils.split_sequence(partitions, npartitions):
            futures: list[dask.distributed.Future] = [
                dask_utils.simple_delayed('insert', _insert)(
                    args=partition,  # type: ignore[arg-type]
                    axis=self.axis,
                    zds=scattered_ds,
                    fs=self.fs,
                    merge_callable=merge_callable,
                    partitioning_properties=self.partition_properties,
                ) for partition in sequence
            ]
            storage.execute_transaction(client, self.synchronizer, futures)

    # pylint: disable=method-hidden
    def drop_partitions(
        self,
        *,
        filters: PartitionFilter = None,
        timedelta: datetime.timedelta | None = None,
    ) -> None:
        # pylint: disable=method-hidden
        """Drop the selected partitions.

        Args:
            filters: The predicate used to filter the partitions to drop. To
                get more information on the predicate, see the documentation of
                the :meth:`partitions` method.
            timedelta: Select the partitions created before the specified time
                delta relative to the current time.

        Example:
            >>> collection.drop_partitions(filters="year == 2019")
            >>> collection.drop_partitions(
            ...     timedelta=datetime.timedelta(days=30))
        """
        now: datetime.datetime = datetime.datetime.now()
        client: dask.distributed.Client = dask_utils.get_client()
        folders = list(self.partitions(filters=filters, lock=True))

        # No partition selected, nothing to do.
        if not folders:
            return

        def is_created_before(path: str, now: datetime.datetime,
                              timedelta: datetime.timedelta) -> bool:
            """Return whether the partition was created before the
            timedelta."""
            created: datetime.datetime = self.fs.created(path)
            return now - created > timedelta

        if timedelta is not None:
            folders = list(
                filter(
                    lambda folder: is_created_before(  # type: ignore[arg-type]
                        folder, now, timedelta),
                    folders))

        storage.execute_transaction(
            client, self.synchronizer,
            client.map(self.fs.rm, folders, recursive=True))

        def invalidate_cache(path) -> None:
            """Invalidate the cache."""
            _LOGGER.info('Dropped partition: %s', path)
            self.fs.invalidate_cache(path)

        tuple(map(invalidate_cache, folders))

    # pylint: disable=method-hidden
    def update(
        self,
        func: UpdateCallable,
        /,
        *args,
        delayed: bool = True,
        depth: int = 0,
        filters: PartitionFilter | None = None,
        npartitions: int | None = None,
        selected_variables: list[str] | None = None,
        trim: bool = True,
        **kwargs,
    ) -> None:
        # pylint: disable=method-hidden
        """Update the selected partitions.

        Args:
            func: The function to apply on each partition.
            *args: The positional arguments to pass to the function.
            delayed: Whether to load data in a dask array or not.
            depth: The depth of the overlap between the partitions. Default is
                0 (no overlap). If depth is greater than 0, the function is
                applied on the partition and its neighbors selected by the
                depth. If ``func`` accepts a partition_info as a keyword
                argument, it will be passed a tuple with the name of the
                partitioned dimension and the slice allowing getting in the
                dataset the selected partition.
            filters: The expression used to filter the partitions to update.
            npartitions: The number of partitions to update in parallel. By
                default, it is equal to the number of Dask workers available
                when calling this method.
            selected_variables: A list of variables to load from the collection.
                If None, all variables are loaded.
            trim: Whether to trim ``depth`` items from each partition after
                calling ``func``. Set it to ``False`` if your function does
                this for you.
            **kwargs: The keyword arguments to pass to the function.

        Raises:
            ValueError: If the variables to update are not in the collection.

        Example:
            >>> import dask.array
            >>> import zcollection
            >>> def ones(ds):
            ...     return dict(var2=ds.variables["var1"].values * 0 + 1)
            >>> collection = zcollection.Collection("my_collection", mode="w")
            >>> collection.update(ones)
        """
        if not callable(func):
            raise TypeError('func must be a callable')

        variables: tuple[str, ...]
        variables = _infer_callable(self, func, filters, delayed,
                                    selected_variables, *args, **kwargs)
        if not variables:
            warnings.warn('You are trying to update an empty collection.',
                          category=RuntimeWarning,
                          stacklevel=2)
            return

        # If depth is not 0, the variables updated must be in the selected
        # variables.
        if depth != 0 and selected_variables is not None:
            selected_variables += list(
                set(variables) - set(selected_variables))

        _LOGGER.info('Updating of the (%s) variable in the collection',
                     ', '.join(repr(item) for item in variables))

        selected_partitions = tuple(self.partitions(filters=filters,
                                                    lock=True))

        local_func: WrappedPartitionCallable = _wrap_update_func(
            *args,
            delayed=delayed,
            func=func,
            fs=self.fs,
            immutable=self._immutable,
            selected_variables=selected_variables,
            **kwargs) if depth == 0 else _wrap_update_func_with_overlap(
                *args,
                delayed=delayed,
                depth=depth,
                dim=self.partition_properties.dim,
                func=func,
                fs=self.fs,
                immutable=self._immutable,
                selected_partitions=selected_partitions,
                selected_variables=selected_variables,
                trim=trim,
                **kwargs)

        client: dask.distributed.Client = dask_utils.get_client()

        batches: Iterator[Sequence[str]] = dask_utils.split_sequence(
            selected_partitions, npartitions
            or dask_utils.dask_workers(client, cores_only=True))
        storage.execute_transaction(
            client, self.synchronizer,
            client.map(local_func, tuple(batches), key=func.__name__))

    def drop_variable(
        self,
        variable: str,
    ) -> None:
        """Delete the variable from the collection.

        Args:
            variable: The variable to delete.

        Raises:
            ValueError: If the variable doesn't exist in the collection or is
                used by the partitioning.

        Example:
            >>> import zcollection
            >>> collection = zcollection.open_collection(
            ...     "my_collection", mode="w")
            >>> collection.drop_variable("my_variable")
        """
        _LOGGER.info('Dropping of the %r variable in the collection', variable)
        if variable in self.partitioning.variables:
            raise ValueError(
                f'The variable {variable!r} is part of the partitioning.')
        if variable not in self.metadata.variables:
            raise ValueError(
                f'The variable {variable!r} does not exist in the collection.')
        if self._immutable:
            zds: dataset.Dataset = storage.open_zarr_group(
                self._immutable, self.fs)
            if variable in zds.variables:
                raise ValueError(
                    f'The variable {variable!r} is part of the immutable '
                    'dataset.')
        client: dask.distributed.Client = dask_utils.get_client()
        bag: dask.bag.core.Bag = self._bag_from_partitions(lock=True)
        awaitables: list[
            dask.distributed.Future] = dask.distributed.futures_of(
                bag.map(storage.del_zarr_array, variable, self.fs).persist())
        storage.execute_transaction(client, self.synchronizer, awaitables)
        del self.metadata.variables[variable]
        self._write_config()

    def add_variable(
        self,
        variable: meta.Variable | dataset.Variable,
    ) -> None:
        """Add a variable to the collection.

        Args:
            variable: The variable to add.

        Raises:
            ValueError: if the variable is already part of the collection, it
                doesn't use the partitioning dimension or use a dimension that
                is not part of the dataset.

        Example:
            >>> import zcollection
            >>> collection = zcollection.open_collection(
            ...     "my_collection", mode="w")
            >>> new_variable = meta.Variable(
            ...     name="my_variable",
            ...     dtype=numpy.dtype("int16"),
            ...     dimensions=("num_lines", "num_pixels"),
            ...     fill_value=32267,
            ...     attrs=(dataset.Attribute(name="my_attr", value=0), ),
            ... )
            >>> collection.add_variable(new_variable)
        """
        variable = dataset.get_variable_metadata(variable)
        _LOGGER.info('Adding of the %r variable in the collection',
                     variable.name)
        if self.partition_properties.dim not in variable.dimensions:
            raise ValueError(
                'The new variable must use the partitioning axis.')
        self.metadata.add_variable(variable)
        self._write_config()

        # Remove the attribute from the variable. The attribute will be added
        # from the collection metadata.
        variable = variable.set_for_insertion()

        client: dask.distributed.Client = dask_utils.get_client()

        template: meta.Variable = self.metadata.search_same_dimensions_as(
            variable)
        chunks: dict[str, int] = {
            dim.name: dim.value
            for dim in self.metadata.chunks
        }
        try:
            bag: dask.bag.core.Bag = self._bag_from_partitions(lock=True)
            futures: list[
                dask.distributed.Future] = dask.distributed.futures_of(
                    bag.map(storage.add_zarr_array,
                            variable,
                            template.name,
                            self.fs,
                            chunks=chunks).persist())
            storage.execute_transaction(client, self.synchronizer, futures)
        except Exception:
            self.drop_variable(variable.name)
            raise

    def copy(
        self,
        target: str,
        *,
        filters: PartitionFilter | None = None,
        filesystem: fsspec.AbstractFileSystem | None = None,
        mode: Literal['r', 'w'] = 'w',
        npartitions: int | None = None,
        synchronizer: sync.Sync | None = None,
    ) -> Collection:
        """Copy the collection to a new location.

        Args:
            target: The target location.
            filters: The predicate used to filter the partitions to copy.
            filesystem: The file system to use. If None, the file system of the
                collection is used.
            mode: The mode used to open the collection copied. Default is 'w'.
            npartitions: The number of partitions top copy in parallel. Default
                is number of cores.
            synchronizer: The synchronizer used to synchronize the collection
                copied. Default is None.

        Returns:
            The new collection.

        Example:
            >>> import zcollection
            >>> collection = zcollection.open_collection(
            ...     "my_collection", mode="r")
            >>> collection.copy(target="my_new_collection")
        """
        _LOGGER.info('Copying of the collection to %r', target)
        if filesystem is None:
            filesystem = fs_utils.get_fs(target)
        client: dask.distributed.Client = dask_utils.get_client()
        npartitions = npartitions or dask_utils.dask_workers(client,
                                                             cores_only=True)

        # Sequence of (source, target) to copy split in npartitions
        args = tuple(
            dask_utils.split_sequence(
                [(item,
                  fs_utils.join_path(
                      target,
                      os.path.relpath(item, self.partition_properties.dir)))
                 for item in self.partitions(filters=filters)], npartitions))
        # Copy the selected partitions
        partial = functools.partial(fs_utils.copy_tree,
                                    fs_source=self.fs,
                                    fs_target=filesystem)

        def worker_task(args: Sequence[tuple[str, str]]) -> None:
            """Function call on each worker to copy the partitions."""
            tuple(map(lambda arg: partial(*arg), args))

        client.gather(client.map(worker_task, args))
        # Then the remaining files in the root directory (config, metadata,
        # etc.)
        fs_utils.copy_files([
            item['name']
            for item in self.fs.listdir(self.partition_properties.dir,
                                        detail=True) if item['type'] == 'file'
        ], target, self.fs, filesystem)
        return Collection.from_config(target,
                                      mode=mode,
                                      filesystem=filesystem,
                                      synchronizer=synchronizer)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Test partitioning by sequence.
==============================
"""
from __future__ import annotations

from typing import Iterator
import pickle

import dask.array.core
import numpy
import pytest
import xarray

from zcollection.type_hints import ArrayLike

from . import data
from .. import Sequence, get_codecs
from ... import dataset
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster

# pylint: enable=unused-import # Need to import for fixtures


def test_construction() -> None:
    """Test the sequence constructor."""
    assert isinstance(Sequence(('a', 'b')), Sequence)
    assert len(Sequence(('a', 'b'))) == 2
    with pytest.raises(ValueError):
        Sequence(('a', 'b'), (0, ))  # type: ignore
    with pytest.raises(ValueError):
        Sequence((), ())
    with pytest.raises(ValueError):
        Sequence(('a', 'b'), dtype=('c', 'd'))
    with pytest.raises(ValueError):
        Sequence(('a', 'b'), dtype=('float32', 'int32'))
    with pytest.raises(TypeError):
        Sequence(('a', 'b'), dtype='int32')
    partitioning = Sequence(('a', 'b'))
    partition_keys = partitioning.parse('a=1/b=2')
    assert partitioning.encode(partition_keys) == (1, 2)
    with pytest.raises(ValueError):
        partitioning.encode((('A', 1), ('b', 2)))
    assert partitioning.decode((1, 2)) == (('a', 1), ('b', 2))
    assert partition_keys == (('a', 1), ('b', 2))
    with pytest.raises(ValueError):
        partitioning.parse('a=1/b=2/c=3')
    with pytest.raises(ValueError):
        partitioning.parse('field=1')


@pytest.mark.parametrize('delayed', [False, True])
def test_split_dataset(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    delayed: bool,
) -> None:
    """Test the split_dataset method."""
    repeatability = 5
    xds = data.create_test_sequence(repeatability, 20, 10)
    partitioning = Sequence(('cycle_number', 'pass_number'))

    cycle_number = 1
    pass_number = 1

    assert partitioning.dtype() == (
        ('cycle_number', 'int64'),
        ('pass_number', 'int64'),
    )

    # Build the test dataset
    zds = dataset.Dataset.from_xarray(xds)
    if not delayed:
        zds = zds.compute()

    iterator = partitioning.split_dataset(zds, 'num_lines')
    assert isinstance(iterator, Iterator)

    for partition, indexer in iterator:
        subset = zds.isel(indexer)
        expected = (f'cycle_number={cycle_number}',
                    f'pass_number={pass_number}')
        assert expected == partition
        assert numpy.all(
            xds.where((xds.cycle_number == cycle_number)
                      & (xds.pass_number == pass_number),
                      drop=True).observation ==
            subset.variables['observation'].array)

        partition_keys = partitioning.parse('/'.join(partition))
        assert partition_keys == (('cycle_number', cycle_number),
                                  ('pass_number', pass_number))
        assert partitioning.decode(
            partitioning.encode(partition_keys)) == partition_keys
        assert partitioning.join(partition_keys, '/') == '/'.join(partition)

        pass_number += 1
        if pass_number > repeatability:
            pass_number = 1
            cycle_number += 1

    xds['cycle_number'] = xarray.DataArray(numpy.array(
        [xds['cycle_number'].values] * 2).T,
                                           dims=('num_lines', 'nump_pixels'))
    zds = dataset.Dataset.from_xarray(xds)
    if not delayed:
        zds = zds.compute()
    with pytest.raises(ValueError):
        list(partitioning.split_dataset(zds, 'num_lines'))


def test_config() -> None:
    """Test the configuration of the Sequence class."""
    partitioning = Sequence(('cycle_number', 'pass_number'))
    config = partitioning.get_config()
    partitioning = get_codecs(config)  # type: ignore[assignment]
    assert isinstance(partitioning, Sequence)


def test_pickle() -> None:
    """Test the pickling of the Date class."""
    partitioning = Sequence(('cycle_number', 'pass_number'))
    other = pickle.loads(pickle.dumps(partitioning))
    assert isinstance(other, Sequence)
    assert other.variables == ('cycle_number', 'pass_number')


# pylint: disable=protected-access
@pytest.mark.parametrize('delayed', [False, True])
def test_multiple_sequence(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    delayed: bool,
) -> None:
    """Test the creation of a sequence with multiple variables."""
    arrays = {
        '_a': numpy.array([], dtype='i8'),
        '_b': numpy.array([], dtype='i8'),
        '_c': numpy.array([], dtype='i8')
    }
    for _a in range(5):
        for _b in range(5):
            arrays['_a'] = numpy.concatenate(
                (arrays['_a'], numpy.full((5, ), _a, dtype='i8')))
            arrays['_b'] = numpy.concatenate(
                (arrays['_b'], numpy.full((5, ), _b, dtype='i8')))
            arrays['_c'] = numpy.concatenate(
                (arrays['_c'], numpy.arange(5, dtype='i8')))
    partitioning = Sequence(('_a', '_b', '_c'))
    chunks: str = (10, )  # type: ignore[assignment]
    if delayed:
        variables: dict[str, ArrayLike] = {  # type: ignore[assignment]
            '_a': dask.array.core.from_array(arrays['_a'], chunks=chunks),
            '_b': dask.array.core.from_array(arrays['_b'], chunks=chunks),
            '_c': dask.array.core.from_array(arrays['_c'], chunks=chunks)
        }
    else:
        variables = arrays  # type: ignore[assignment]
    _a = 0
    _b = 0
    _c = 0
    for idx, item in enumerate(
            partitioning._split(variables)):  # type: ignore[arg-type]
        assert item[0] == (('_a', _a), ('_b', _b), ('_c', _c))
        _c += 1
        if _c > 4:
            _c = 0
            _b += 1
        if _b > 4:
            _b = 0
            _a += 1
        assert item[1] == slice(idx, idx + 1)

    numpy.random.shuffle(arrays['_c'])
    variables['_c'] = dask.array.core.from_array(  # type: ignore[assignment]
        arrays['_c'], chunks=chunks) if delayed else arrays['_c']

    with pytest.raises(ValueError):
        list(partitioning._split(variables))  # type: ignore[arg-type]

    del variables['_c']
    del variables['_b']
    partitioning = Sequence(('_a', '_b', '_c'))

    _a = 0
    for idx, item in enumerate(
            partitioning._split(variables)):  # type: ignore[arg-type]
        assert item[0] == (('_a', _a), )
        _a += 1
        assert item[1] == slice(idx * 25, idx * 25 + 25)
    # pylint: enable=protected-access


def test_values_must_be_integer() -> None:
    """Test that the values must be integer."""
    values = numpy.arange(0, 100, dtype='f8')
    partitioning = Sequence(('values', ))
    # pylint: disable=protected-access
    with pytest.raises(TypeError):
        list(partitioning._split({'values': values}))
    # pylint: enable=protected-access



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Fixture for testing the file system.
====================================
"""
from typing import Any, Iterator
import pathlib

import fsspec
import fsspec.implementations.memory
import pytest

try:
    # pylint: disable=unused-import # Need to import for fixtures
    from .s3 import S3, s3, s3_base  # type: ignore

    # pylint: disable=unused-import
    S3_IMPORT_EXCEPTION = None
except ImportError as err:
    S3_IMPORT_EXCEPTION = str(err)


class Local:
    """Local files system."""

    def __init__(self, tmpdir, protocol) -> None:
        #: The filesystem.
        self.fs: fsspec.AbstractFileSystem = fsspec.filesystem(protocol)
        #: The root directory.
        self.root = pathlib.Path(tmpdir)
        #: The collection directory.
        self.collection: pathlib.Path = self.root.joinpath('collection')
        #: The view directory.
        self.view: pathlib.Path = self.root.joinpath('view')

    def __getattr__(self, name) -> Any:
        return getattr(self.fs, name)


@pytest.fixture
def local_fs(tmpdir, pytestconfig) -> Iterator[Local]:
    """Local filesystem."""
    protocol: str = 'memory' if pytestconfig.getoption('memory') else 'file'
    instance = Local(tmpdir, protocol)
    yield instance
    try:
        # For the memory protocol we delete the written data to free the
        # memory.
        if isinstance(instance.fs,
                      fsspec.implementations.memory.MemoryFileSystem):
            instance.fs.rm(str(instance.root), recursive=True)
    except FileNotFoundError:
        pass


# pylint: disable=redefined-outer-name,function-redefined
if S3_IMPORT_EXCEPTION is None:

    @pytest.fixture
    def s3_fs(s3) -> S3:  # type: ignore[arg-type]
        """S3 filesystem."""
        return S3(s3)  # type: ignore
else:

    @pytest.fixture
    def s3() -> None:
        """S3 filesystem."""

    @pytest.fixture
    def s3_base() -> None:
        """S3 filesystem."""

    @pytest.fixture
    def s3_fs(pytestconfig) -> None:
        """S3 filesystem."""
        if pytestconfig.getoption('s3'):
            pytest.fail(f'Unable to test S3: {S3_IMPORT_EXCEPTION}')
        else:
            pytest.skip('S3 is disabled')


# pylint: enable=redefined-outer-name,function-redefined



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Test of views
=============
"""
from __future__ import annotations

import pathlib

import numpy
import pytest

from ... import collection, convenience, meta, partitioning, view
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster
from ...tests.data import (
    create_test_collection,
    create_test_dataset,
    make_dataset,
)
from ...tests.fixture import dask_arrays, numpy_arrays
from ...tests.fs import local_fs, s3, s3_base, s3_fs
from ...view.detail import _calculate_axis_reference

# pylint: enable=unused-import


@pytest.mark.parametrize('fs', ['local_fs', 's3_fs'])
@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_view(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    fs,
    arrays_type,
    request,
):
    """Test the creation of a view."""
    tested_fs = request.getfixturevalue(fs)
    delayed = request.getfixturevalue(arrays_type)

    create_test_collection(tested_fs, delayed=False)
    instance = convenience.create_view(str(tested_fs.view),
                                       view.ViewReference(
                                           str(tested_fs.collection),
                                           tested_fs.fs),
                                       filesystem=tested_fs.fs)
    assert isinstance(instance, view.View)
    assert isinstance(str(instance), str)

    # No variable recorded, so no data can be loaded
    with pytest.raises(ValueError):
        instance.load(delayed=delayed)

    var = meta.Variable(
        name='var2',
        dtype=numpy.float64,
        dimensions=('num_lines', 'num_pixels'),
        attrs=(meta.Attribute(name='attr', value=1), ),
    )

    with pytest.raises(ValueError):
        instance.add_variable(var)

    var.name = 'var3'
    instance.add_variable(var)

    with pytest.raises(ValueError):
        instance.add_variable(var)

    instance = convenience.open_view(str(tested_fs.view),
                                     filesystem=tested_fs.fs)
    zds = instance.load(delayed=delayed)
    assert zds is not None
    assert set(zds['time'].values.astype('datetime64[D]')) == {
        numpy.datetime64('2000-01-01'),
        numpy.datetime64('2000-01-04'),
        numpy.datetime64('2000-01-07'),
        numpy.datetime64('2000-01-10'),
        numpy.datetime64('2000-01-13'),
        numpy.datetime64('2000-01-16'),
    }

    # Loading a variable existing only in the view.
    zds = instance.load(delayed=delayed, selected_variables=('var3', ))
    assert zds is not None
    assert tuple(zds.variables) == ('var3', )
    assert 'var3' in zds.metadata().variables.keys()

    # The metadata of the reference collection is not modified.
    assert 'var3' not in instance.view_ref.metadata.variables.keys()

    # Loading a non existing variable.
    zds = instance.load(delayed=delayed, selected_variables=('var55', ))
    assert zds is not None
    assert len(zds.variables) == 0

    # Test view loading that is no longer synchronized with the reference
    # collection.
    tested_fs.fs.rm(str(
        tested_fs.view.joinpath('year=2000', 'month=01', 'day=13')),
                    recursive=True)

    assert len(tuple(instance.partitions())) == 5
    assert len(tuple(instance.view_ref.partitions())) == 6

    zds = instance.load(delayed=delayed)
    assert zds is not None
    assert set(zds['time'].values.astype('datetime64[D]')) == {
        numpy.datetime64('2000-01-01'),
        numpy.datetime64('2000-01-04'),
        numpy.datetime64('2000-01-07'),
        numpy.datetime64('2000-01-10'),
        numpy.datetime64('2000-01-16'),
    }

    # Create a variable with the unsynchronized view
    var.name = 'var4'
    instance.add_variable(var)

    zds = instance.load(delayed=delayed)
    assert zds is not None

    def update(zds, varname):
        """Update function used for this test."""
        return {varname: zds.variables['var1'].values * 0 + 5}

    instance.update(update, 'var3', delayed=delayed)  # type: ignore

    with pytest.raises(ValueError):
        instance.update(update, 'varX')  # type: ignore

    with pytest.raises(ValueError):
        instance.update(update, 'var2')  # type: ignore

    zds = instance.load(delayed=delayed)
    assert zds is not None
    numpy.all(zds.variables['var3'].values == 5)

    indexers = instance.map(
        lambda x: slice(0, x.dimensions['num_lines'])  # type: ignore
    ).compute()
    ds1 = instance.load(delayed=delayed, indexer=indexers)
    assert ds1 is not None
    ds2 = instance.load(delayed=delayed)
    assert ds2 is not None

    assert numpy.allclose(ds1.variables['var1'].values,
                          ds2.variables['var1'].values)

    instance.drop_variable('var3')

    assert tuple(
        str(pathlib.Path(item))
        for item in instance.partitions(filters=instance.sync())) == (str(
            tested_fs.view.joinpath('year=2000', 'month=01', 'day=13')), )

    with pytest.raises(ValueError):
        convenience.open_view(str(tested_fs.collection),
                              filesystem=tested_fs.fs)


@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_view_overlap(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arg,
    request,
):
    """Test the creation of a view."""
    tested_fs = request.getfixturevalue(arg)

    create_test_collection(tested_fs)
    instance = convenience.create_view(str(tested_fs.view),
                                       view.ViewReference(
                                           str(tested_fs.collection),
                                           tested_fs.fs),
                                       filesystem=tested_fs.fs)

    var = meta.Variable(
        name='var3',
        dtype=numpy.int8,
        dimensions=('num_lines', 'num_pixels'),
    )

    instance.add_variable(var)

    def update(zds, varname, partition_info: tuple[str, slice],
               trim_result: bool) -> dict[str, numpy.ndarray]:
        """Update function used for this test."""
        assert isinstance(partition_info, tuple)
        assert len(partition_info) == 2
        assert isinstance(partition_info[0], str)
        assert isinstance(partition_info[1], slice)
        assert partition_info[0] == 'num_lines'
        if trim_result:
            zds = zds.isel(dict((partition_info, )))
        return {varname: zds.variables['var1'].values * 1 + 5}

    instance.update(
        update,  # type: ignore
        'var3',
        depth=1,
        trim_result=False)
    zds = instance.load()
    assert zds is not None
    numpy.all(zds.variables['var3'].values == 5)

    instance.update(
        update,  # type: ignore
        'var3',
        depth=1,
        trim=False,
        trim_result=True)
    zds = instance.load()
    assert zds is not None
    numpy.all(zds.variables['var3'].values == 5)

    with pytest.raises(ValueError):
        instance.update(
            update,  #  type: ignore
            'var3',
            depth=1,
            trim_result=False,
            selected_variables=('var1', ))

    def map_func(_, partition_info: tuple[str, slice]):
        """Map function used for this test."""
        assert isinstance(partition_info, tuple)
        assert len(partition_info) == 2
        assert isinstance(partition_info[0], str)
        assert isinstance(partition_info[1], slice)
        assert partition_info[0] == 'num_lines'
        return partition_info

    indexers = instance.map_overlap(
        map_func,  # type: ignore
        depth=1,
    ).compute()

    for _, data in indexers:
        assert isinstance(data, tuple)
        assert len(data) == 2
        assert isinstance(data[0], str)
        assert isinstance(data[1], slice)


def test_view_checksum(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
        tmpdir) -> None:
    """Test the checksum calculation."""
    zds = next(create_test_dataset())
    zcollection = collection.Collection('time', zds.metadata(),
                                        partitioning.Date(('time', ), 'D'),
                                        str(tmpdir))

    zcollection.insert(zds)
    partition = tmpdir / 'year=2000' / 'month=01' / 'day=01'
    axis_ref = _calculate_axis_reference(str(partition), zcollection)
    assert isinstance(axis_ref.array, numpy.ndarray)
    assert isinstance(axis_ref.checksum, str)
    assert isinstance(axis_ref.dimension, str)
    assert len(axis_ref.checksum) == 64
    assert axis_ref.dimension == 'num_lines'
    assert axis_ref.checksum == ('1a1727b18e729c376442e76d806565b8'
                                 '359e3af9c93572af3e5fe8980ced6956')


@pytest.mark.filterwarnings('ignore:.*cannot be serialized.*')
@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_view_sync(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arg,
    request,
):
    """Test the synchronization of a view."""
    tested_fs = request.getfixturevalue(arg)
    create_test_collection(tested_fs)
    instance = convenience.create_view(str(tested_fs.view),
                                       view.ViewReference(
                                           str(tested_fs.collection),
                                           tested_fs.fs),
                                       filesystem=tested_fs.fs)
    var = meta.Variable(name='var3',
                        dtype=numpy.float64,
                        dimensions=('num_lines', 'num_pixels'))
    instance.add_variable(var)
    del instance

    zcollection = convenience.open_collection(str(tested_fs.collection),
                                              filesystem=tested_fs.fs,
                                              mode='w')
    zds = zcollection.load(filters=lambda keys: keys['year'] == 2000 and keys[
        'month'] == 1 and keys['day'] == 16)
    assert zds is not None
    dates = numpy.arange(numpy.datetime64('2000-01-16'),
                         numpy.datetime64('2000-01-16T23:59:59'),
                         numpy.timedelta64(1, 'h'))
    zds = make_dataset(
        dates.astype('M8[ns]'),
        numpy.ones((len(dates), zds.dimensions['num_pixels']),
                   dtype=numpy.int64))
    zcollection.insert(zds)
    del zcollection
    instance = convenience.open_view(str(tested_fs.view),
                                     filesystem=tested_fs.fs)
    assert instance is not None
    assert instance.is_synced() is False
    instance.sync(filters=lambda keys: True)
    zds = instance.load()
    assert zds is not None



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Partitioning a sequence of variables
====================================
"""
from __future__ import annotations

from typing import Any, ClassVar, Iterator

import dask.array.core
import dask.array.routines
import numpy

from . import abc
from ..type_hints import ArrayLike, NDArray


def _is_monotonic(arr: NDArray) -> bool:
    """Check if the array is monotonic.

    The matrix will be sorted in the reverse order of the partitioning keys
    (column in the matrix). If the order of the matrix is unchanged, the
    different partitioning columns are monotonic.

    Args:
        arr: The array to check.

    Returns:
        True if the array is monotonic, False otherwise.
    """
    # `reversed` because `numpy.lexsort` wants the most significant key last.
    values: list[NDArray] = [
        arr[:, ix] for ix in reversed(range(arr.shape[1]))
    ]
    sort_order: NDArray = numpy.lexsort(numpy.array(values))
    return numpy.all(abc.difference(sort_order) > 0)  # type: ignore


def _unique(arr: ArrayLike, is_delayed: bool) -> tuple[NDArray, NDArray]:
    """Return unique elements and their indices.

    Args:
        arr: Array of elements.
        is_delayed: If True, the array is delayed.
    Returns:
        Tuple of unique elements and their indices.
    """
    index: NDArray
    indices: NDArray

    if is_delayed:
        index, indices = abc.unique(arr)  # type: ignore[arg-type]
        if not _is_monotonic(index):
            raise ValueError('index is not monotonic')
        return index, indices
    return abc.unique_and_check_monotony(arr)


class Sequence(abc.Partitioning):
    """Initialize a partitioning scheme for a sequence of variables.

    A sequence is a combination of variables constituting unique monotonic keys.
    For example, the orbit number (``cycle``) and the half-orbit number
    (``pass``) of a satellite.

    Args:
        variables: A list of strings representing the variables to be used for
            partitioning.
        dtype: An optional sequence of strings representing the data type used
            to store variable values in a binary representation without data
            loss. Must be one of the following allowed data types: ``int8``,
            ``int16``, ``int32``, ``int64``, ``uint8``, ``uint16``, ``uint32``,
            ``uint64``. If not provided, defaults to ``int64`` for all
            variables.

    Raises:
        ValueError: If the periodicity is not valid.

    Example:
        >>> partitioning = Sequence(["a", "b", "c"], (None, 10, 10))
    """
    #: The ID of the partitioning scheme.
    ID: ClassVar[str] = 'Sequence'

    # pylint: disable=arguments-differ
    # False positive: `self` is used in the signature.
    @staticmethod
    def _split(variables: dict[str, ArrayLike]) -> Iterator[abc.Partition]:
        """Split the variables constituting the partitioning into partitioning
        schemes."""
        index: NDArray
        indices: NDArray
        matrix: dask.array.core.Array | NDArray

        # Determine if the variables are handled by Dask.
        is_delayed: bool = any(
            isinstance(item, dask.array.core.Array)
            for item in variables.values())

        # Combines the arrays of variable values into a transposed matrix.
        matrix = dask.array.routines.vstack(tuple(
            variables.values())).transpose() if is_delayed else numpy.vstack(
                tuple(variables.values())).transpose()
        if matrix.dtype.kind not in 'iu':
            raise TypeError('The variables must be integer')

        index, indices = _unique(matrix, is_delayed)  # type: ignore[arg-type]
        indices = abc.concatenate_item(indices, matrix.shape[0])

        fields = tuple(variables.keys())
        # pylint: disable=unnecessary-lambda-assignment
        # We want to reference a lambda function, not assign it to a variable.
        if len(fields) == 1:
            concat: Any = lambda fields, keys: (fields + keys, )
        else:
            concat = lambda fields, keys: tuple(zip(fields, keys))
        # pylint: enable=unnecessary-lambda-assignment

        return ((concat(fields,
                        tuple(item)), slice(start, indices[ix + 1], None))
                for item, (ix, start) in zip(index, enumerate(indices[:-1])))
        # pylint: enable=arguments-differ

    def encode(
        self,
        partition: tuple[tuple[str, int], ...],
    ) -> tuple[int, ...]:
        """Encode a partitioning scheme to the handled values.

        Args:
            partition: The partitioning scheme to be encoded.

        Returns:
            The encoded partitioning scheme.

        Example:
            >>> partitioning = Sequence(["a", "b", "c"])
            >>> fields = partitioning.parse("a=100/b=10/c=1")
            >>> fields
            (('a', 100), ('b', 10), ('c', 1))
            >>> partitioning.encode(fields)
            (100, 10, 1)
        """
        return tuple(value
                     for _, value in self.parse(self.join(partition, '/')))

    def decode(self, values: tuple[int, ...]) -> tuple[tuple[str, int], ...]:
        """Decode a partitioning scheme.

        Args:
            values: The encoded partitioning scheme.

        Returns:
            The decoded partitioning scheme.

        Example:
            >>> partitioning = Sequence(["a", "b", "c"])
            >>> partitioning.decode((100, 10, 1))
            (('a', 100), ('b', 10), ('c', 1))
        """
        return tuple(
            (key, value) for key, value in zip(self.variables, values))



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Dataset variable.
=================
"""
from __future__ import annotations

from typing import Any, Callable, Iterator, Sequence, TypeVar
import abc
import collections

import dask.array.core
import numcodecs.abc
import numpy
import xarray
import zarr

from .. import mathematics, meta, representation
from ..meta import Attribute
from ..type_hints import ArrayLike, NDArray, NDMaskedArray

#: Generic type for a variable interface.
I = TypeVar('I', bound='IVariable')

#: Generic type for a dataset variable.
T = TypeVar('T', bound='Variable')


def new_variable(cls: type[I], **kwargs: Any) -> I:
    """Create a new variable.

    Args:
        constructor: Variable constructor.
        kwargs: Keyword arguments passed to the constructor.
    """
    self: I = cls.__new__(cls)
    # pylint: disable=expression-not-assigned
    # We use the set notation to evaluate the generator
    {
        setattr(self, key, value)  # type: ignore[func-returns-value]
        for key, value in kwargs.items()
    }
    # pylint: enable=expression-not-assigned
    return self


def concat(
    self: T,
    other: T | Sequence[T],
    concatenate: Callable,
    dim: str,
) -> T:
    """Concatenate a variable with another variable or a sequence of variables.

    Args:
        self: A variable.
        other: A variable or a sequence of variables.
        constructor: Variable constructor.
        concatenate: Function used to concatenate the arrays.
        dim: Dimension along which the arrays are concatenated.

    Returns:
        A new variable.

    Raises:
        ValueError: If ``other`` is empty or is not a sequence of ``self`` type.
    """
    if not isinstance(other, Sequence):
        other = [other]
    if not other:
        raise ValueError('other must be a non-empty sequence')

    # Self and other must have the same type
    if not all(isinstance(item, type(self)) for item in other):
        raise ValueError(f'other must be a sequence of {type(self)}')

    try:
        axis: int = self.dimensions.index(dim)
        return new_variable(
            type(self),
            name=self.name,
            array=concatenate((self.array, *(item.array for item in other)),
                              axis=axis),
            dimensions=self.dimensions,
            attrs=self.attrs,
            compressor=self.compressor,
            fill_value=self.fill_value,
            filters=self.filters,
        )
    except ValueError:
        # If the concatenation dimension is not within the dimensions of the
        # variable, then the original variable is returned (i.e.
        # concatenation is not necessary).
        return new_variable(type(self),
                            name=self.name,
                            array=self.array,
                            dimensions=self.dimensions,
                            attrs=self.attrs,
                            compressor=self.compressor,
                            fill_value=self.fill_value,
                            filters=self.filters)


def not_equal(first: Any, second: Any) -> bool:
    """Check if two objects are not equal.

    Args:
        first: The first object.
        second: The second object.

    Returns:
        True if the two objects are different, False otherwise.
    """

    def _is_not_a_number(number: Any) -> bool:
        """Check if a number is NaN or NaT."""
        # pylint: disable=comparison-with-itself
        return not number == number and number != number
        # pylint: enable=comparison-with-itself

    #: pylint: disable=unidiomatic-typecheck
    if type(first) != type(second):
        return True
    #: pylint: enable=unidiomatic-typecheck
    if _is_not_a_number(first) and _is_not_a_number(second):
        return False
    if first is None and second is None:
        return False
    if first == second:
        return False
    return True


def _variable_repr(var: Variable) -> str:
    """Get the string representation of a variable.

    Args:
        var: A variable.

    Returns:
        The string representation of the variable.
    """
    # Dimensions
    dims_str: str = representation.dimensions(
        dict(zip(var.dimensions, var.shape)))
    lines: list[str] = [
        f'<{var.__module__}.{var.__class__.__name__} {dims_str}>',
        f'{var.array!r}'
    ]
    # Attributes
    if len(var.attrs):
        lines.append('  Attributes:')
        lines += representation.attributes(var.attrs)
    # Filters
    if var.filters:
        lines.append('  Filters:')
        lines += [f'    {item!r}' for item in var.filters]
    # Compressor
    if var.compressor:
        lines.append('  Compressor:')
        lines += [f'    {var.compressor!r}']
    return '\n'.join(lines)


class IVariable(abc.ABC):
    """Define the interface for a variable."""

    @property
    @abc.abstractmethod
    def data(self) -> dask.array.core.Array:
        """Return the values as a dask array where values equal to the fill
        value are masked. If no fill value is set, the returned array is the
        same as the underlying array.

        Returns:
            The dask array

        .. seealso::

            :meth:`Variable.array`
        """

    @property
    @abc.abstractmethod
    def values(self) -> NDArray | NDMaskedArray:
        """Return the variable values.

        .. note::

            If the variable has a fill value, the result is a masked array where
            masked values are equal to the fill value.

        Returns:
            The variable data
        """

    @values.setter
    @abc.abstractmethod
    def values(self, data: Any) -> None:
        """Defines the values array. If the data provided is a masked array,
        it's converted to an array, where the masked values are replaced by the
        fill value of this instance.

        Args:
            data: The new data to use

        Raises:
            ValueError: If the shape of the data does not match the shape of
                the stored data.
        """

    @abc.abstractmethod
    def persist(self: I, **kwargs) -> I:
        """Persist the variable data into memory.

        Args:
            **kwargs: Additional arguments to pass to the persist method.

        Returns:
            The variable
        """

    @abc.abstractmethod
    def compute(self, **kwargs) -> NDArray | NDMaskedArray:
        """Return the variable data as a numpy array.

        .. note::

            If the variable has a fill value, the result is a masked array where
            masked values are equal to the fill value.

        Args:
            **kwargs: Additional arguments to pass to the compute method.
        """

    @abc.abstractmethod
    def fill(self: I) -> I:
        """Fill the variable with the fill value. If the variable has no fill
        value, this method does nothing.

        Returns:
            The variable.
        """

    @classmethod
    @abc.abstractmethod
    def from_zarr(cls: type[I], array: zarr.Array, name: str, dimension: str,
                  **kwargs) -> I:
        """Create a new variable from a zarr array.

        Args:
            array: The zarr array
            name: Name of the variable
            dimension: Name of the attribute that defines the dimensions of the
                variable
            **kwargs: Additional arguments to pass to the method.

        Returns:
            The variable
        """

    @abc.abstractmethod
    def concat(self: I, other: I | Sequence[I], dim: str) -> I:
        """Concatenate this variable with another variable or a list of
        variables along a dimension.

        Args:
            other: Variable or list of variables to concatenate with this
                variable.
            dim: Dimension to concatenate along.

        Returns:
            New variable.

        Raises:
            ValueError: if the variables provided is an empty sequence.
        """

    @abc.abstractmethod
    def __getitem__(self, key: Any) -> Any:
        """Get a slice of the variable.

        Args:
            key: Slice or index to use.
        Returns:
            The variable slice.
        """

    @abc.abstractmethod
    def rechunk(self: I, **kwargs) -> I:
        """Rechunk the variable.

        Args:
            **kwargs: Keyword arguments passed to
                :func:`dask.array.rechunk`

        Returns:
            The variable.
        """


class Variable(IVariable):
    """Variables hold multi-dimensional arrays of data.

    Args:
        name: Name of the variable
        data: Variable data
        dimensions: Variable dimensions
        attrs: Variable attributes
        compressor: Compression codec
        fill_value: Value to use for uninitialized values
        filters: Filters to apply before writing data to disk
    """
    __slots__ = ('array', 'attrs', 'compressor', 'dimensions', 'fill_value',
                 'filters', 'name')

    def __init__(self,
                 name: str,
                 data: Any,
                 dimensions: Sequence[str],
                 *,
                 attrs: Sequence[Attribute] | None = None,
                 compressor: numcodecs.abc.Codec | None = None,
                 fill_value: Any | None = None,
                 filters: Sequence[numcodecs.abc.Codec] | None = None) -> None:
        #: Variable name
        self.name: str = name
        #: Variable data as a dask array.
        self.array: Any = data
        #: Variable dimensions
        self.dimensions = tuple(dimensions)
        #: Variable attributes
        self.attrs = tuple(attrs or ())
        #: Compressor used to compress the data during writing data to disk
        self.compressor: numcodecs.abc.Codec | None = compressor
        #: Value to use for uninitialized values
        self.fill_value: Any | None = fill_value
        #: Filters to apply before writing data to disk
        self.filters = tuple(filters or ())

    @property
    def dtype(self) -> numpy.dtype:
        """Return the dtype of the underlying array."""
        return self.array.dtype

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the variable."""
        return self.array.shape

    @property
    def ndim(self) -> int:
        """Return the number of dimensions of the variable."""
        return len(self.dimensions)

    @property
    def size(self: Any) -> int:
        """Return the size of the variable."""
        return mathematics.prod(self.shape)

    @property
    def nbytes(self) -> int:
        """Return the number of bytes used by the variable."""
        return self.size * self.dtype.itemsize

    def duplicate(self: T, data: Any) -> T:
        """Create a new variable from the properties of this instance and the
        data provided.

        Args:
            data: Variable data.

        Returns:
            New variable.

        Raises:
            ValueError: If the shape of the data does not match the shape of
                the stored data.
        """
        result: T = type(self)(self.name,
                               data,
                               self.dimensions,
                               attrs=self.attrs,
                               compressor=self.compressor,
                               fill_value=self.fill_value,
                               filters=self.filters)
        if len(result.shape) != len(self.dimensions):
            raise ValueError('data shape does not match variable dimensions')
        return result

    def isel(self: T, key: tuple[slice, ...]) -> T:
        """Return a new variable with data selected along the given dimension
        indices.

        Args:
            key: Dimension indices to select

        Returns:
            The new variable
        """
        return new_variable(type(self),
                            name=self.name,
                            array=self.array[key],
                            dimensions=self.dimensions,
                            attrs=self.attrs,
                            compressor=self.compressor,
                            fill_value=self.fill_value,
                            filters=self.filters)

    def set_for_insertion(self: T) -> T:
        """Create a new variable without any attribute.

        Returns:
            The variable.
        """
        return new_variable(type(self),
                            name=self.name,
                            array=self.array,
                            dimensions=self.dimensions,
                            attrs=tuple(),
                            compressor=self.compressor,
                            fill_value=self.fill_value,
                            filters=self.filters)

    def rename(self: T, name: str) -> T:
        """Rename the variable.

        Args:
            name: New variable name.

        Returns:
            The variable.
        """
        return new_variable(type(self),
                            name=name,
                            array=self.array,
                            dimensions=self.dimensions,
                            attrs=self.attrs,
                            compressor=self.compressor,
                            fill_value=self.fill_value,
                            filters=self.filters)

    def fill_attrs(self, var: meta.Variable) -> None:
        """Fill the variable attributes using the provided metadata.

        Args:
            var: Variable's metadata.
        """
        self.attrs = var.attrs

    def metadata(self) -> meta.Variable:
        """Get the variable metadata.

        Returns:
            Variable metadata
        """
        return meta.Variable(self.name,
                             self.dtype,
                             dimensions=self.dimensions,
                             attrs=self.attrs,
                             compressor=self.compressor,
                             fill_value=self.fill_value,
                             filters=self.filters)

    def have_same_properties(self, other: Variable) -> bool:
        """Return true if this instance and the other variable have the same
        properties."""
        return self.metadata() == other.metadata()

    def dimension_index(self) -> Iterator[tuple[str, int]]:
        """Return an iterator over the variable dimensions and their index.

        Returns:
            An iterator over the variable dimensions
        """
        yield from ((item, ix) for ix, item in enumerate(self.dimensions))

    def to_xarray(self) -> xarray.Variable:
        """Convert the variable to an xarray.Variable.

        Returns:
            Variable as an xarray.Variable
        """
        encoding: dict[str, Any] = {}
        if self.filters:
            encoding['filters'] = self.filters
        if self.compressor:
            encoding['compressor'] = self.compressor
        data: ArrayLike = self.array
        if self.dtype.kind == 'M':
            # xarray need a datetime64[ns] dtype
            data = data.astype('datetime64[ns]')
            encoding['dtype'] = 'int64'
        elif self.dtype.kind == 'm':
            encoding['dtype'] = 'int64'
        attrs = collections.OrderedDict(
            (item.name, item.value) for item in self.attrs)
        if self.fill_value is not None:
            attrs['_FillValue'] = self.fill_value
        return xarray.Variable(self.dimensions, data, attrs, encoding)

    def __str__(self) -> str:
        return _variable_repr(self)

    def __repr__(self) -> str:
        return _variable_repr(self)

    def __hash__(self) -> int:
        return hash(self.name)

    def __array__(self) -> ArrayLike:
        return self.values



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Dask utilities
==============
"""
from __future__ import annotations

from typing import Any, Callable, Iterator, Sequence
import itertools
import uuid

from dask.delayed import Delayed as dask_Delayed
import dask.distributed
import dask.highlevelgraph


def dask_workers(client: dask.distributed.Client,
                 cores_only: bool = False) -> int:
    """Return the number of dask workers available.

    Args:
        client: dask client
        cores_only: if True, only the number of cores is returned,
            otherwise the total number of threads is returned.

    Returns:
        number of dask workers

    Raises:
        ValueError: If no dask workers are available.
    """
    result: int = len(
        client.ncores()) if cores_only else sum(  # type: ignore[arg-type]
            item
            for item in client.nthreads().values())  # type: ignore[arg-type]
    if result == 0:
        raise RuntimeError('No dask workers available')
    return result


def get_client() -> dask.distributed.Client:
    """Return the default dask client.

    Returns:
        default dask client
    """
    try:
        return dask.distributed.get_client()
    except ValueError:
        return dask.distributed.Client(
            processes=False,
            direct_to_workers=True,
        )


def split_sequence(sequence: Sequence[Any],
                   sections: int | None = None) -> Iterator[Sequence[Any]]:
    """Split a sequence into sections.

    Args:
        sequence: The sequence to split.
        sections: The number of sections to split the sequence into. Default
            divides the sequence into n sections of one element.

    Returns:
        Iterator of sequences.
    """
    sections = len(sequence) if sections is None else sections
    if sections <= 0:
        raise ValueError('The number of sections must be greater than zero.')
    length: int = len(sequence)
    sections = min(sections, length)

    size: int
    extras: int
    size, extras = divmod(length, sections)

    div = tuple(
        itertools.accumulate([0] + extras * [size + 1] +
                             (sections - extras) * [size]))
    yield from (sequence[item:div[ix + 1]] for ix, item in enumerate(div[:-1]))


def simple_delayed(name: str, func: Callable) -> dask_Delayed:
    """Create a simple delayed function.

    Args:
        name: name of the function
        func: function to be delayed

    Returns:
        delayed function
    """
    name = f'{name}-{str(uuid.uuid4())}'
    return dask_Delayed(
        name,
        dask.highlevelgraph.HighLevelGraph({name: {
            name: func
        }}, {name: set()}),
        None,
    )



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Indexing a Collection.
======================
"""
import warnings

try:
    from .abc import Indexer, QueryDict, Scalar
    __all__ = ('Indexer', 'QueryDict', 'Scalar')
except ImportError:  # pragma: no cover
    warnings.warn(
        'Install PyArrow to use the indexing capabilities of zcollection.')



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Mathematics testing.
====================
"""
from .. import mathematics


def test_prod():
    """Test the product of an iterable."""
    assert mathematics.prod([]) == 1
    assert mathematics.prod([1]) == 1
    assert mathematics.prod([1, 2, 3]) == 6
    assert mathematics.prod([1, 2, 3, 4, 5]) == 120
    assert mathematics.prod([1, 2, 3, 4, 5, 6, 7, 8, 9]) == 362880
    assert mathematics.prod([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]) == 3628800



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Synchronization of concurrent accesses
======================================
"""
from __future__ import annotations

from typing import Callable
import abc
import threading

import fasteners


class Sync(abc.ABC):  # pragma: no cover
    """Interface of the classes handling the synchronization of concurrent
    accesses."""

    @abc.abstractmethod
    def __enter__(self) -> bool:
        ...

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...

    @abc.abstractmethod
    def is_locked(self) -> bool:
        """Returns True if the lock is acquired, False otherwise."""


class NoSync(Sync):
    """This class is used when the user does not want to synchronize accesses
    to the collection, in other words, when there is no concurrency."""

    def __enter__(self) -> bool:
        return True

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """As this class does not perform any synchronization, this method has
        nothing to do."""

    def is_locked(self) -> bool:
        """As this class does not perform any synchronization, this method
        always returns False."""
        return False


class ProcessSync(Sync):
    """This class is used when the user wants to synchronize accesses to the
    collection, in other words, when there is concurrency."""

    def __init__(self, path: str) -> None:
        self.lock = fasteners.InterProcessLock(path)

    def __enter__(self) -> bool:
        return self.lock.acquire()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        try:
            self.lock.release()
        except threading.ThreadError:
            pass

    def __reduce__(self) -> tuple[Callable, tuple[str]]:
        return (ProcessSync, (str(self.lock.path), ))

    def is_locked(self) -> bool:
        """Returns True if the lock is acquired, False otherwise."""
        return self.lock.exists()



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Partitioning scheme.
====================

Entry point of the implemented partitioning schemes.

* :py:class:`Sequence <zcollection.partitioning.sequence.Sequence>`:
  Partitioning a sequence of variables.
* :py:class:`Date <zcollection.partitioning.date.Date>`: Partitioning a
  sequence of dates.

.. class:: Partitioning

    Alias for :class:`zcollection.partitioning.abc.Partitioning`.
"""
from .abc import Partitioning
from .date import Date
from .registry import get_codecs, register_codec
from .sequence import Sequence

register_codec(Date)
register_codec(Sequence)

__all__ = ('Partitioning', 'Date', 'Sequence', 'get_codecs')



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Make test datasets
==================
"""
from __future__ import annotations

from typing import Iterator
import itertools

import numpy
import zarr

from .. import collection, dataset, partitioning
from ..type_hints import NDArray

#: First date of the test dataset.
START_DATE = numpy.datetime64('2000-01-01', 'ns')
#: Last date of the test dataset.
END_DATE = numpy.datetime64('2000-06-30', 'ns')
#: Delta between two dates.
DELTA = numpy.timedelta64(72, 'h')
#: Fill value.
FILL_VALUE = 2147483647


def make_dataset(dates: numpy.ndarray,
                 measures: numpy.ndarray,
                 fill_value: float | None = None,
                 filters: tuple | None = None,
                 delayed: bool = True) -> dataset.Dataset:
    """Create a dataset."""
    array_class = (dataset.DelayedArray if delayed else dataset.Array)
    return dataset.Dataset(
        attrs=(dataset.Attribute(name='attr', value=1), ),
        variables=(
            array_class(
                name='time',
                data=dates,
                dimensions=('num_lines', ),
                attrs=(dataset.Attribute(name='attr', value=1), ),
                compressor=zarr.Blosc(),
            ),
            array_class(name='var1',
                        data=measures,
                        dimensions=('num_lines', 'num_pixels'),
                        attrs=(dataset.Attribute(name='attr', value=1), ),
                        fill_value=fill_value,
                        filters=filters),
            array_class(name='var2',
                        data=measures,
                        dimensions=('num_lines', 'num_pixels'),
                        attrs=(dataset.Attribute(name='attr', value=1), ),
                        fill_value=fill_value,
                        filters=filters),
        ),
    )


def create_test_dataset(delayed: bool = True) -> Iterator[dataset.Dataset]:
    """Create a temporal dataset."""

    dates: NDArray = numpy.arange(START_DATE, END_DATE, DELTA)
    indices: NDArray = numpy.arange(0, len(dates))

    for item in numpy.array_split(dates, 12):
        mask: NDArray = (dates >= item[0]) & (dates <= item[-1])
        measures = numpy.vstack((indices[mask], ) * 25).T

        yield make_dataset(item, measures, delayed=delayed)


def create_test_dataset_with_fillvalue(
        delayed: bool = True) -> Iterator[dataset.Dataset]:
    """Create a dataset with a fixed scale offset filter and fill values."""

    dates: NDArray = numpy.arange(START_DATE, END_DATE, DELTA)
    measures: NDArray = numpy.arange(0, len(dates), dtype=numpy.float64)
    measures[measures % 2 == 0] = FILL_VALUE
    measures = numpy.vstack((measures, ) * 25).T * 1e-4

    yield make_dataset(dates,
                       measures,
                       delayed=delayed,
                       fill_value=FILL_VALUE * 1e-4,
                       filters=(zarr.FixedScaleOffset(scale=10000,
                                                      offset=0,
                                                      dtype='<f8',
                                                      astype='i4'), ))


def create_test_collection(tested_fs,
                           with_fillvalue=False,
                           delayed=True) -> collection.Collection:
    """Create a collection."""
    zds: dataset.Dataset = next(
        create_test_dataset_with_fillvalue(
            delayed=delayed) if with_fillvalue else create_test_dataset(
                delayed=delayed))
    zcollection = collection.Collection('time',
                                        zds.metadata(),
                                        partitioning.Date(('time', ), 'D'),
                                        str(tested_fs.collection),
                                        filesystem=tested_fs.fs)
    zcollection.insert(zds)
    return zcollection


#: List of filesystems and datasets to test.
FILE_SYSTEM_DATASET = list(
    itertools.product([
        'local_fs',
        's3_fs',
    ], [
        create_test_dataset,
        create_test_dataset_with_fillvalue,
    ]))



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Dataset
=======
"""
from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    OrderedDict,
    Sequence,
    Tuple,
    Union,
)
import collections

import dask.array.core
import dask.array.routines
import dask.array.wrap
import dask.base
import numpy
import xarray
import xarray.conventions

from . import meta, representation
from .compressed_array import CompressedArray
from .meta import BLOCK_SIZE_LIMIT, Attribute, Dimension
from .type_hints import ArrayLike, NDArray, NDMaskedArray
from .variable import Array, DelayedArray, Variable, new_variable

#: Alias to type hint for the dimensions of a dataset.
DimensionType = Dict[str, int]

#: Alias to type hint for the variables of a dataset.
VariableType = OrderedDict[str, Variable]

#: Alias to type hint for the attributes of a dataset.
AttributeType = Tuple[Attribute, ...]

#: Alias to type hint for the chunk sizes for each dimension of a dataset.
ChunkType = Dict[str, Union[int, str]]


def _dask_repr(array: dask.array.core.Array) -> str:
    """Get the string representation of a dask array.

    Args:
        array: A dask array.

    Returns:
        The string representation of the dask array.
    """
    chunksize = tuple(item[0] for item in array.chunks)
    return f'dask.array<chunksize={chunksize}>'


def _dataset_repr(zds: Dataset) -> str:
    """Get the string representation of a dataset.

    Args:
        zds: A dataset.

    Returns:
        The string representation of the dataset.
    """
    # Dimensions
    dims_str: str = representation.dimensions(zds.dimensions)
    lines: list[str] = [
        f'<{zds.__module__}.{zds.__class__.__name__}>',
        f'  Dimensions: {dims_str}', 'Data variables:'
    ]
    # Variables
    if len(zds.variables) == 0:
        lines.append('    <empty>')
    else:
        width: int = representation.calculate_column_width(zds.variables)
        for name, var in zds.variables.items():
            dims_str = f"({', '.join(map(str, var.dimensions))}) "
            name_str: str = f'    {name:<{width}s} {dims_str} {var.dtype}'
            data_str: str = _dask_repr(var.data) if zds.delayed else '...'
            lines.append(
                representation.pretty_print(f'{name_str}: {data_str}'))
    # Attributes
    if len(zds.attrs):
        lines.append('  Attributes:')
        lines += representation.attributes(zds.attrs)

    return '\n'.join(lines)


def _duplicate_delayed_array(var: DelayedArray,
                             data: dask.array.core.Array) -> DelayedArray:
    """Duplicate the variable with a new data.

    Args:
        var: Variable to duplicate.
        data: The new data to use

    Returns:
        The duplicated variable
    """
    return new_variable(type(var),
                        name=var.name,
                        array=data,
                        dimensions=var.dimensions,
                        attrs=var.attrs,
                        compressor=var.compressor,
                        fill_value=var.fill_value,
                        filters=var.filters)


def _duplicate_array(var: Array, data: NDArray) -> Array:
    """Duplicate the variable with a new data.

    Args:
        var: Variable to duplicate.
        data: The new data to use

    Returns:
        The duplicated variable
    """
    return new_variable(type(var),
                        name=var.name,
                        array=data,
                        dimensions=var.dimensions,
                        attrs=var.attrs,
                        compressor=var.compressor,
                        fill_value=var.fill_value,
                        filters=var.filters)


def _delete_delayed_vars(self: Dataset, indexer: slice | Sequence[int],
                         axis: str) -> list[Variable]:
    """Delete the variables along the given axis."""
    return [
        _duplicate_delayed_array(
            var,  # type: ignore[arg-type]
            dask.array.routines.delete(var.array, indexer,
                                       var.dimensions.index(axis)))
        for var in self.variables.values()
    ]


def _delete_vars(self: Dataset, indexer: slice | Sequence[int],
                 axis: str) -> list[Variable]:
    """Delete the variables along the given axis."""
    return [
        _duplicate_array(
            var,  # type: ignore[arg-type]
            numpy.delete(var.array, indexer, var.dimensions.index(axis)))
        for var in self.variables.values()
    ]


def _update_dimensions(self: Dataset,
                       delayed: bool | None = None) -> bool | None:
    """Update the dimensions of the dataset based on its variables."""
    for var in self.variables.values():
        if delayed is None:
            delayed = isinstance(var, DelayedArray)
        elif delayed != isinstance(var, DelayedArray):
            raise ValueError(
                'the dataset contains both delayed and non-delayed '
                'variables')
        try:
            for idx, dim in enumerate(var.dimensions):
                if dim not in self.dimensions:
                    self.dimensions[dim] = var.array.shape[idx]
                elif self.dimensions[dim] != var.array.shape[idx]:
                    raise ValueError(f'variable {var.name} has conflicting '
                                     'dimensions')
        except IndexError as exc:
            raise ValueError(
                f'variable {var.name} has missing dimensions') from exc
    return delayed


class Dataset:
    """Hold variables, dimensions, and attributes that together form a dataset.

    Args:
        variables (DelayedArray | Array): A dictionary of variables in the
            dataset, with variable names as keys and :py:class:`Array
            <zcollection.variable.array.Array>` or :py:class:`DelayedArray
            <zcollection.variable.delayed.DelayedArray>` objects as values.
        attrs: A tuple of global attributes on this dataset.
        block_size_limit: The maximum size (in bytes) of a block/chunk of
            variable's data. Defaults to 128 MiB.
        chunks: A dictionary of chunk sizes for each dimension.
        delayed: A boolean indicating whether the dataset contains delayed
            variables (numpy arrays wrapped in dask arrays).

    Raises:
        ValueError: If the dataset contains variables with the same dimensions
            but with different values.
        ValueError: If the dataset contains both delayed and non-delayed
            variables.

    Notes:
        The dataset is a dictionary-like container of variables. It also holds
        the dimensions and attributes of the dataset.
        If the dataset contains delayed variables, the values are
        :py:class:`DelayedArray
        <zcollection.variable.delayed_array.DelayedArray>` objects. Otherwise,
        the values are :py:class:`Array
        <zcollection.variable.array.Array>` objects. It is impossible to mix
        delayed and non-delayed variables in the same dataset.
    """
    __slots__ = ('dimensions', 'variables', 'attrs', 'chunks',
                 'block_size_limit', 'delayed')

    def __init__(self,
                 variables: Iterable[Variable],
                 *,
                 attrs: Sequence[Attribute] | None = None,
                 block_size_limit: int | None = None,
                 chunks: Sequence[Dimension] | None = None,
                 delayed: bool | None = None) -> None:
        #: The list of global attributes on this dataset
        self.attrs = tuple(attrs or ())

        #: Dataset contents as dict of :py:class:`Variable
        #: <zcollection.variable.abc.Variable>` objects. If the dataset
        #: contains delayed variables, the values are :py:class:`DelayedArray
        #: <zcollection.variable.delayed_array.DelayedArray>` objects.
        #: Otherwise, the values are :py:class:`Array
        #: <zcollection.variable.array.Array>` objects.
        self.variables = collections.OrderedDict(
            (item.name, item) for item in variables)

        #: A dictionary of dimension names and their index in the dataset
        self.dimensions: DimensionType = {}

        # Loops over each variable in the dataset and updates the dimensions
        # according to the shape of the array.
        delayed = _update_dimensions(self, delayed)

        #: Maximum data chunk size
        self.block_size_limit: int = block_size_limit or BLOCK_SIZE_LIMIT

        #: Chunk size for each dimension
        self.chunks: ChunkType = {dim.name: dim.value for dim in chunks or []}

        #: The type of variables in the dataset
        self.delayed: bool = delayed if delayed is not None else True

    def __len__(self) -> int:
        return len(self.variables)

    def __bool__(self) -> bool:
        return bool(self.variables)

    def __getitem__(self, name: str) -> Variable:
        """Return a variable from the dataset.

        Args:
            name: The name of the variable to return

        Returns:
            The variable

        Raises:
            KeyError: If the variable is not found
        """
        return self.variables[name]

    def __getattr__(self, name: str) -> Any:
        if name in self.variables:
            return self.variables[name]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __getstate__(self) -> tuple[Any, ...]:
        return (self.dimensions, self.variables, self.attrs, self.chunks,
                self.block_size_limit, self.delayed)

    def __setstate__(
        self, state: tuple[DimensionType, VariableType, AttributeType,
                           ChunkType, int, bool]
    ) -> None:
        (self.dimensions, self.variables, self.attrs, self.chunks,
         self.block_size_limit, self.delayed) = state

    @property
    def nbytes(self) -> int:
        """Return the total number of bytes in the dataset.

        Returns:
            The total number of bytes in the dataset
        """
        return sum(item.nbytes for item in self.variables.values())

    @property
    def dims_chunk(self) -> Sequence[Dimension]:
        """Dimensions chunk size as a tuple.

        Returns:
            Dimensions associated to their chunk size.
        """
        return tuple(Dimension(*item) for item in self.chunks.items())

    def add_variable(self,
                     var: meta.Variable,
                     /,
                     data: ArrayLike[Any] | None = None) -> None:
        """Add a variable to the dataset.

        Args:
            var: The variable to add
            data: The data to add to the variable. If not provided, the variable
                will be created with the default fill value.

        Raises:
            ValueError: If the variable added has dimensions that conflict with
                existing dimensions, or if the variable has dimensions not
                defined in the dataset.
        """
        if set(var.dimensions) - set(self.dimensions):
            raise ValueError(f'variable {var.name} has dimensions '
                             f'{var.dimensions} that are not in the dataset')

        if data is None:
            shape = tuple(self.dimensions[dim] for dim in var.dimensions)
            data = dask.array.wrap.full(
                shape, var.fill_value,
                dtype=var.dtype) if self.delayed else numpy.full(
                    shape, var.fill_value, dtype=var.dtype)
        else:
            for dim, size in zip(var.dimensions, data.shape):
                if size != self.dimensions[dim]:
                    raise ValueError(
                        f'Conflicting sizes for dimension {dim!r}: '
                        f'length {self.dimensions[dim]} on the data but length '
                        f'{size} defined in dataset.')
        self.variables[var.name] = (DelayedArray if self.delayed else Array)(
            var.name,
            data,
            var.dimensions,
            attrs=var.attrs,
            compressor=var.compressor,
            fill_value=var.fill_value,
            filters=var.filters,
        )

    def rename(self, names: Mapping[str, str]) -> None:
        """Rename variables in the dataset.

        Args:
            names: A mapping from old names to new names

        Raises:
            ValueError: If the new names conflict with existing names
        """
        for old, new in names.items():
            if new in self.variables:
                raise ValueError(f'{new} already exists in the dataset')
            self.variables[new] = self.variables.pop(old).rename(new)

    def drops_vars(self, names: str | Sequence[str]) -> None:
        """Drop variables from the dataset.

        Args:
            names: Variable names to drop.
        """
        if isinstance(names, str) or not isinstance(names, Iterable):
            names = [names]
        tuple(map(self.variables.pop, names))

    def select_vars(self, names: str | Sequence[str]) -> Dataset:
        """Return a new dataset containing only the selected variables.

        Args:
            names: Variable names to select.

        Returns:
            A new dataset containing only the selected variables.
        """
        if isinstance(names, str) or not isinstance(names, Iterable):
            names = [names]
        return Dataset((self.variables[name] for name in names),
                       attrs=self.attrs,
                       block_size_limit=self.block_size_limit,
                       chunks=self.dims_chunk)

    def metadata(self) -> meta.Dataset:
        """Get the dataset metadata.

        Returns:
            Dataset metadata
        """
        return meta.Dataset(
            dimensions=tuple(self.dimensions.keys()),
            variables=tuple(item.metadata()
                            for item in self.variables.values()),
            attrs=self.attrs,
            chunks=tuple(Dimension(*item) for item in self.chunks.items()),
            block_size_limit=self.block_size_limit,
        )

    @staticmethod
    def from_xarray(zds: xarray.Dataset, delayed: bool = True) -> Dataset:
        """Create a new dataset from a xarray dataset.

        Args:
            zds: Dataset to convert.
            delayed: If True, the data will be wrapped in a dask array. If
                False, the data will be handled as a numpy array.

        Returns:
            New dataset.
        """
        handler: type[DelayedArray | Array]
        handler = DelayedArray if delayed else Array
        variables: list[Variable] = [
            handler(
                name,  # type: ignore[arg-type]
                array.data,  # type: ignore[arg-type]
                tuple(array.dims),  # type: ignore[arg-type]
                attrs=tuple(
                    Attribute(*attr)  # type: ignore[arg-type]
                    for attr in array.attrs.items()),
                compressor=array.encoding.get('compressor', None),
                fill_value=array.encoding.get('_FillValue', None),
                filters=array.encoding.get('filters', None))
            for name, array in zds.variables.items()
        ]

        return Dataset(
            variables=variables,
            attrs=tuple(
                Attribute(*item)  # type: ignore[arg-type]
                for item in zds.attrs.items()),
            delayed=delayed,
        )

    def to_xarray(self, **kwargs) -> xarray.Dataset:
        """Convert the dataset to a xarray dataset.

        Args:
            **kwargs: Additional parameters are passed through the function
                :py:func:`xarray.conventions.decode_cf_variables`.

        Returns:
            Dataset as a xarray dataset.
        """
        data_vars = collections.OrderedDict(
            (name, var.to_xarray()) for name, var in self.variables.items())
        attrs = collections.OrderedDict(
            (item.name, item.value) for item in self.attrs)
        data_vars, attrs, coord_names = xarray.conventions.decode_cf_variables(
            data_vars, attrs, **kwargs)
        result = xarray.Dataset(data_vars, attrs=attrs)
        return result.set_coords(coord_names.intersection(data_vars))

    def to_dict(
        self,
        variables: Sequence[str] | None = None,
        **kwargs,
    ) -> dict[str, NDArray | NDMaskedArray]:
        """Convert the dataset to a dictionary, between the variable names and
        their data.

        Args:
            variables: Variables to include (default to all dataset's
                variables).
            **kwargs: Additional parameters are passed through the function
                :py:func:`dask.compute`.

        Returns:
            Dictionary of variables.
        """
        variables = variables or tuple(self.variables.keys())
        if self.delayed:
            arrays = tuple((key, value.data)
                           for key, value in self.variables.items()
                           if key in variables)
            return dict(dask.base.compute(*arrays, **kwargs))
        return dict(
            tuple((key, value.values) for key, value in self.variables.items()
                  if key in variables))

    def set_for_insertion(self, mds: meta.Dataset) -> Dataset:
        """Create a new dataset ready to be inserted into a collection.

        Args:
            mds: Dataset metadata.

        Returns:
            New dataset.
        """
        return Dataset(
            variables=[
                var.set_for_insertion() for var in self.variables.values()
            ],
            chunks=mds.chunks,
            block_size_limit=mds.block_size_limit,
        )

    def fill_attrs(self, mds: meta.Dataset) -> None:
        """Fill the dataset and its variables attributes using the provided
        metadata.

        Args:
            mds: Dataset metadata.
        """
        self.attrs = tuple(mds.attrs)
        tuple(
            map(lambda var: var.fill_attrs(mds.variables[var.name]),
                self.variables.values()))

    def isel(self, slices: dict[str, Any]) -> Dataset:
        """Return a new dataset with each array indexed along the specified
        slices.

        Args:
            slices: Dictionary of dimension names and slices

        Returns:
            New dataset.
        """
        dims_invalid: set[str] = set(slices) - set(self.dimensions)
        if dims_invalid:
            raise ValueError(
                f'Slices contain invalid dimension name(s): {dims_invalid}')
        variables: list[Variable] = [
            var.isel(
                tuple(slices.get(dim, slice(None)) for dim in var.dimensions))
            for var in self.variables.values()
        ]
        return Dataset(variables=variables,
                       attrs=self.attrs,
                       block_size_limit=self.block_size_limit,
                       chunks=self.dims_chunk,
                       delayed=self.delayed)

    def delete(self, indexer: slice | Sequence[int], axis: str) -> Dataset:
        """Return a new dataset without the data selected by the provided
        indices.

        Args:
            indexer: Indices to remove along the specified axis.
            axis: The axis along which to delete the subarrays defined
                in the dataset.

        Returns:
            New dataset.
        """
        variables: list[Variable] = _delete_delayed_vars(
            self, indexer, axis) if self.delayed else _delete_vars(
                self, indexer, axis)
        return Dataset(variables=variables,
                       attrs=self.attrs,
                       block_size_limit=self.block_size_limit,
                       chunks=self.dims_chunk,
                       delayed=self.delayed)

    def compute(self, **kwargs) -> Dataset:
        """Compute the dataset variables.

        Args:
            **kwargs: Additional parameters are passed through to
                :py:func:`dask.array.compute`.

        Returns:
            New dataset.
        """
        if not self.delayed:
            return self

        arrays: Iterable[NDArray] = dask.base.compute(
            *tuple(item.array for item in self.variables.values()), **kwargs)

        variables: list[Variable] = [
            Array(item.name,
                  array,
                  item.dimensions,
                  attrs=item.attrs,
                  compressor=item.compressor,
                  fill_value=item.fill_value,
                  filters=item.filters)
            for item, array in zip(self.variables.values(), arrays)
        ]
        return Dataset(variables=variables,
                       attrs=self.attrs,
                       chunks=self.dims_chunk,
                       block_size_limit=self.block_size_limit,
                       delayed=False)

    def rechunk(self, **kwargs) -> Dataset:
        """Rechunk the dataset.

        Args:
            **kwargs: Keyword arguments are passed through to
                :py:func:`dask.array.rechunk.rechunk`.

        Returns:
            New dataset.

        .. seealso:: :py:func:`dask.array.rechunk`
        """
        if not self.delayed:
            return self
        variables: list[Variable] = [
            var.rechunk(**kwargs) for var in self.variables.values()
        ]
        return Dataset(variables=variables,
                       attrs=self.attrs,
                       chunks=self.dims_chunk,
                       block_size_limit=self.block_size_limit)

    def persist(
        self,
        *,
        compress: bool = False,
        **kwargs,
    ) -> Dataset:
        """Persist the dataset variables.

        Args:
            compress: If true, compress the data loaded into memory.
            **kwargs: Additional parameters are passed to the function
                :py:func:`dask.array.Array.persist`.

        Returns:
            The dataset with the variables persisted into memory.
        """
        if not self.delayed:
            return self
        if compress:
            for var in self.variables.values():
                var.array = var.array.map_blocks(CompressedArray,
                                                 fill_value=var.fill_value)
        arrays: Iterable[NDArray] = dask.base.persist(
            *tuple(item.data for item in self.variables.values()), **kwargs)
        variables: OrderedDict[str, Variable] = self.variables
        for name, array in zip(self.variables, arrays):
            variables[name].array = array

        return self

    def concat(self, other: Dataset | Iterable[Dataset], dim: str) -> Dataset:
        """Concatenate datasets along a dimension.

        Args:
            other: Datasets to concatenate.
            dim: Dimension along which to concatenate the datasets.

        Returns:
            New dataset.

        Raises:
            ValueError: If the provided sequence of datasets is empty.
        """
        if not isinstance(other, Iterable):
            other = [other]
        other = tuple(other)
        if not other:
            raise ValueError('cannot concatenate an empty sequence')
        if not all(item.delayed == self.delayed for item in other):
            raise ValueError('cannot concatenate delayed and non-delayed data')
        variables: list[Variable] = [
            var.concat(tuple(item.variables[name] for item in other), dim)
            for name, var in self.variables.items()
        ]
        return Dataset(variables=variables,
                       attrs=self.attrs,
                       block_size_limit=self.block_size_limit,
                       chunks=self.dims_chunk,
                       delayed=self.delayed)

    def merge(self, other: Dataset) -> None:
        """Merge the provided dataset into this dataset.

        Args:
            other: Dataset to merge into this dataset.
        """
        if self.delayed != other.delayed:
            raise ValueError('cannot merge delayed and non-delayed data')

        # Merge the variables
        for name, var in other.variables.items():

            # It's impossible to merge a variable with itself.
            if name in self.variables:
                raise ValueError(f'variable {name} already exists')
            self.variables[name] = var

        # If the dataset has common dimensions, they must be identical.
        same_dims: set[str] = set(self.dimensions) & set(other.dimensions)
        if same_dims and not all(self.dimensions[dim] == other.dimensions[dim]
                                 for dim in same_dims):
            raise ValueError(f'dimensions {same_dims} are not identical')

        # Merge the dimensions.
        self.dimensions.update(other.dimensions)

        # Merge the attributes (overwriting any existing attributes).
        if self.attrs is None:
            self.attrs = other.attrs
        elif other.attrs is not None:
            attrs = dict(item.get_config() for item in self.attrs)
            attrs.update(item.get_config() for item in other.attrs)
            self.attrs = tuple(Attribute(*item) for item in attrs.items())

    def select_variables_by_dims(self,
                                 dims: Sequence[str],
                                 predicate: bool = True) -> Dataset:
        """Return a new dataset with only the variables that have the specified
        dimensions if predicate is true, otherwise return a new dataset with
        only the variables that do not have the specified dimensions.

        Args:
            dims: Dimensions to select.
            predicate: If true, select variables with the specified dimensions,
                otherwise select variables without the specified dimensions.

        Returns:
            New dataset or None if no variables match the predicate.
        """

        def _predicate_for_dimension_less(var: Variable) -> bool:
            """Return true if the variable is selected by the predicate."""
            return (len(var.dimensions) == 0) == predicate

        def _predicate_for_dimension(var: Variable) -> bool:
            """Return true if the variable is selected by the predicate."""
            return bool(set(var.dimensions) & set_of_dims) == predicate

        condition: Callable[..., bool]
        condition = (_predicate_for_dimension_less
                     if not dims else _predicate_for_dimension)

        set_of_dims = set(dims)
        variables: list[Variable] = [
            var for var in self.variables.values() if condition(var)
        ]
        return Dataset(variables=variables,
                       attrs=self.attrs,
                       chunks=self.dims_chunk,
                       block_size_limit=self.block_size_limit)

    # def to_zarr(self,
    #             path: str,
    #             fs: fsspec.AbstractFileSystem | None = None,
    #             parallel: bool = True) -> None:
    #     """Write the dataset to a Zarr store.

    #     Args:
    #         path: Path to the Zarr store.
    #         fs: Filesystem to use.
    #         parallel: If true, write the data in parallel.
    #     """
    #     if not self.delayed:
    #         raise ValueError('cannot write a non-delayed dataset to Zarr')
    #     # pylint: disable=import-outside-toplevel, import-error
    #     # Avoid circular import
    #     import storage
    #     import sync

    #     # pylint: enable=import-outside-toplevel, import-error
    #     storage.write_zarr_group(self, path, fs or fsspec.filesystem('file'),
    #                              sync.NoSync(), parallel)

    def __str__(self) -> str:
        return _dataset_repr(self)

    def __repr__(self) -> str:
        return _dataset_repr(self)


def get_variable_metadata(var: Variable | meta.Variable) -> meta.Variable:
    """Get the variable metadata.

    Args:
        var: Variable to get the metadata for.

    Returns:
        Variable metadata.
    """
    if isinstance(var, Variable):
        return var.metadata()
    return var


def get_dataset_variable_properties(
        metadata: meta.Dataset,
        selected_variables: Iterable[str] | None = None) -> tuple[Array, ...]:
    """Return the variables properties defined in the dataset.

    Args:
        metadata: Metadata dataset containing variables information.
        selected_variables: The variables to return. If None, all the
            variables are returned.

    Returns:
        The variables defined in the dataset.
    """
    selected_variables = selected_variables or metadata.variables.keys()
    return tuple(
        Array(v.name,
              numpy.ndarray((0, ) * len(v.dimensions), v.dtype),
              v.dimensions,
              attrs=v.attrs,
              compressor=v.compressor,
              fill_value=v.fill_value,
              filters=v.filters) for k, v in metadata.variables.items()
        if k in selected_variables)



"""
Overview of a View.
===================

This section outlines the steps required to get started with the main features
of a ``View``.
"""
from typing import Iterator
import pprint

import dask.distributed
import fsspec
import numpy

import zcollection
import zcollection.tests.data


# %%
# Initialization of the environment
# ---------------------------------
#
# As in the example of handling
# :ref:`collections <sphx_glr_auto_examples_ex_collection.py>`, we will create
# the test environment and a collection.
def create_dataset() -> zcollection.Dataset:
    """Create a dataset to record."""
    generator: Iterator[zcollection.Dataset] = \
        zcollection.tests.data.create_test_dataset_with_fillvalue()
    return next(generator)


cluster = dask.distributed.LocalCluster(processes=False)
client = dask.distributed.Client(cluster)

zds: zcollection.Dataset | None = create_dataset()
assert zds is not None
fs: fsspec.AbstractFileSystem = fsspec.filesystem('memory')
collection: zcollection.Collection = zcollection.create_collection(
    'time',
    zds,
    zcollection.partitioning.Date(('time', ), resolution='M'),
    '/view_reference',
    filesystem=fs)
collection.insert(zds, merge_callable=zcollection.merging.merge_time_series)

# %%
# Creation of views
# -----------------
#
# A :py:class:`view<zcollection.view.View>` allows you to extend a collection
# (:py:class:`a view reference<zcollection.view.ViewReference>`) that you are
# not allowed to modify.
view: zcollection.View = zcollection.create_view(
    '/my_view',
    zcollection.view.ViewReference('/view_reference', fs),
    filesystem=fs)

# %%
# .. note::
#
#     The created view can be accessed using the following command ::
#
#         >>> view = zcollection.open_view("/my_view", filesystem=fs)
#
# Editing variables
# -----------------
# When the view is created, it has no data of its own, it uses all the
# partitions defined in the reference view. You can select the partitions used
# from the reference collection by specifying the keyword argument ``filters``
# during the creation of the view.
pprint.pprint(fs.listdir('/my_view'))

# %%
# It's not yet possible to read data from the view, as it does not yet have any
# data. To minimize the risk of mismatches with the reference view, the data
# present in the view drives the range of data that can be read.
try:
    view.load()
except ValueError as err:
    print(err)

# %%
# Such a state of the view is not very interesting. But it is possible to
# :py:meth:`add<zcollection.view.View.add_variable>` and modify variables in
# order to enhance the view.
var3_template: zcollection.meta.Variable = zds.metadata().variables['var2']
var3_template.name = 'var3'
view.add_variable(var3_template)
del var3_template

# %%
# This step creates all necessary partitions for the new variable.
pprint.pprint(fs.listdir('/my_view/year=2000'))

# %%
# The new variable is not initialized.
zds = view.load()
assert zds is not None
zds.variables['var3'].values

# %%
# The same principle used by the collection allows to
# :py:meth:`update<zcollection.view.View.update>` the variables.
view.update(
    lambda ds: dict(var3=ds['var1'].values * 0 + 1))  # type: ignore[arg-type]

# %%
# Like the :py:meth:`update<zcollection.collection.Collection.update>` method
# of the collection, the update method of view allows to selecting the
# neighboring partitions with the keyword argument ``depth``.

# %%
zds = view.load()
assert zds is not None
var3: numpy.ndarray = zds['var3'].values
print(var3)

# %%
# **Warning**: The variables of the reference collection cannot be edited.
try:
    view.update(
        lambda ds: dict(var2=ds['var2'].values * 0))  # type: ignore[arg-type]
except ValueError as exc:
    print(str(exc))


# %%
# Sync the view with the reference
# --------------------------------
# The view may not be read anymore if the number of elements in the reference
# collection and in the view is not identical. To avoid this problem, the view
# is automatically synchronized when it is opened. But only if the reference
# collection has been completed (adding new data after the existing data), the
# data already present in the view are kept. The existing tables in the view are
# resized and filled with the defined fill values. If you want to know which
# partitions are synchronized, you have to use the following data flow: open the
# view and ask not to synchronize it (``resync=False``), then call the ``sync``
# method of view class to obtain a filter allowing selecting all the partitions
# that have been modified.
#
# Let's illustrate this data flow with an example.
#
# First, we create an utility function to resize a dataset.
def resize(ds: zcollection.Dataset, dim: str,
           size: int) -> zcollection.Dataset:
    """Resize a dataset."""

    def new_shape(
        var: zcollection.Variable,
        selected_dim: str,
        new_size: int,
    ) -> tuple[int, ...]:
        """Compute the new shape of a variable."""
        return tuple(new_size if dim == selected_dim else size
                     for dim, size in zip(var.dimensions, var.shape))

    return zcollection.Dataset([
        zcollection.Array(
            name,
            numpy.resize(var.array.compute(), new_shape(var, dim, size)),
            var.dimensions,
            attrs=var.attrs,
            compressor=var.compressor,
            fill_value=var.fill_value,
            filters=var.filters,
        ) for name, var in ds.variables.items()
    ])


# %%
# We then modify the last partition of the reference collection. We start by
# opening the reference collection and loading the last partition.
collection = zcollection.open_collection('/view_reference',
                                         filesystem=fs,
                                         mode='w')
zds = collection.load(
    filters=lambda keys: keys['month'] == 6 and keys['year'] == 2000)
assert zds is not None

# %%
# We create a new time variable, resize the dataset and insert the new time
# values.
time: numpy.ndarray = numpy.arange(
    numpy.datetime64('2000-06-01T00:00:00'),
    numpy.datetime64('2000-06-30T23:59:59'),
    numpy.timedelta64(1, 'h'),
)
zds = resize(zds, 'num_lines', len(time))
zds['time'].values = time

# %%
# Finally, we update the partition in the reference collection.
collection.insert(zds)

# %%
# Now we cannot load the view, because the shape of the last partition is no
# longer consistent between the reference collection and the view.
try:
    view.load()
except ValueError as err:
    print(err)

# %%
# We call the ``sync`` method to resynchronize the view.
filters = view.sync()

# %%
# The method returns a callable that can be used to filter the partitions that
# have been synchronized. You can use this information to perform a
# :py:meth:`update<zcollection.view.View.update>` of the view on the
# synchronized partitions: ::
#
#     view.update(
#         lambda ds: dict(var3=ds['var1'].values * 0 + 1),
#         filters=filters)
#
print(tuple(view.partitions(filters=filters)))

# %%
# The view is now synchronized and can be loaded.
zds = view.load()
assert zds is not None
zds.variables['var3'].values

# %%
# Map a function over the view
# ----------------------------
# It's possible to map a function over the partitions of the view.
for partition, array in view.map(lambda ds: (  # type: ignore[arg-type]
        ds['var1'].values + ds['var2'].values)).compute():
    print(f' * partition = {partition}: mean = {array.mean()}')

# %%
# .. seealso::
#
#     See the :py:meth:`map_overlap<zcollection.view.View.map_overlap>` method
#     apply a function over the partitions of the view of selecting the
#     neighboring partitions.
#
# Drop a variable
# ----------------
# A method allows you to
# :py:meth:`drop_variable<zcollection.view.View.drop_variable>` variables from
# the view.
view.drop_variable('var3')
try:
    view.load()
except ValueError as err:
    # The view no longer has a variable.
    print(err)

# %%
# **Warning**: The variables of the reference collection cannot be dropped.
try:
    view.drop_variable('var2')
except ValueError as exc:
    print(str(exc))

# %%
# Close the local cluster to avoid printing warning messages in the other
# examples.
client.close()
cluster.close()



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Metadata testing.
=================
"""
from __future__ import annotations

from typing import Any
import json
import pathlib
import pickle

import numpy
import pytest
import zarr.codecs

from .. import meta


def test_attribute() -> None:
    """Test attribute creation."""
    att = meta.Attribute('a', 23.4)
    assert isinstance(att, meta.Attribute)
    assert att.name == 'a'
    assert att.value == 23.4
    assert str(att) == "Attribute('a', 23.4)"
    # pylint: disable=comparison-with-itself
    assert att == att
    assert (att == 'X') is False
    assert att != meta.Attribute('a', '23.4')
    assert isinstance(meta.Attribute.from_config(att.get_config()),
                      meta.Attribute)

    att = meta.Attribute('a', numpy.arange(10))
    assert att == meta.Attribute('a', numpy.arange(10))

    att = meta.Attribute('a', numpy.datetime64('2000-01-01', 'us'))
    assert att == att
    # pylint: enable=comparison-with-itself


def test_dimension() -> None:
    """Test dimension creation."""
    dim = meta.Dimension('a', 12)
    assert isinstance(dim, meta.Dimension)
    assert dim.name == 'a'
    assert dim.value == 12
    assert str(dim) == "Dimension('a', 12)"
    # pylint: disable=comparison-with-itself
    assert dim == dim
    # pylint: enable=comparison-with-itself
    assert dim != meta.Dimension('a', 11)
    assert isinstance(meta.Dimension.from_config(dim.get_config()),
                      meta.Dimension)


def test_variable() -> None:
    """Test variable creation."""
    var = meta.Variable('a',
                        numpy.dtype('int16'),
                        dimensions=('a', ),
                        attrs=(meta.Attribute('x', 12), ),
                        compressor=zarr.codecs.Zlib(),
                        filters=(zarr.codecs.Delta(numpy.float64, numpy.int16),
                                 zarr.codecs.FixedScaleOffset(
                                     0, 1, numpy.int16)))
    assert isinstance(var, meta.Variable)
    assert str(var) == "Variable('a')"
    # pylint: disable=comparison-with-itself
    assert var == var
    assert (var == 2) is False
    other: meta.Variable = meta.Variable.from_config(var.get_config())
    assert var == other
    other.name = 'x'
    assert var != other
    # pylint: enable=comparison-with-itself


def test_dataset() -> None:
    """Test dataset creation."""
    root: pathlib.Path = pathlib.Path(__file__).parent
    with root.joinpath('first_dataset.json').open(encoding='utf-8') as stream:
        first: dict[str, Any] = json.load(stream)
    with root.joinpath('second_dataset.json').open(encoding='utf-8') as stream:
        second: dict[str, Any] = json.load(stream)
    ds: meta.Dataset = meta.Dataset.from_config(first)
    other: meta.Dataset = meta.Dataset.from_config(second)
    assert ds == other
    assert (ds == 2) is False
    assert (ds != other) is False
    ds.dimensions = ds.dimensions + ('dummy', )
    assert ds != other


def test_select_variables() -> None:
    """Test select_variables."""
    root: pathlib.Path = pathlib.Path(__file__).parent
    with root.joinpath('first_dataset.json').open(encoding='utf-8') as stream:
        config: dict[str, Any] = json.load(stream)
    ds: meta.Dataset = meta.Dataset.from_config(config)
    variables: set[str] = ds.select_variables(('longitude', 'latitude'))
    assert variables == {'longitude', 'latitude'}
    variables = ds.select_variables(drop_variables=('longitude', 'latitude'))
    assert set(variables) & {'longitude', 'latitude'} == set()
    variables = ds.select_variables(keep_variables=('longitude', 'latitude',
                                                    'time'),
                                    drop_variables=('time', ))
    assert variables == {'longitude', 'latitude'}


def test_search_same_dimensions_as() -> None:
    """Test search_same_dimensions_as."""
    root: pathlib.Path = pathlib.Path(__file__).parent
    with root.joinpath('first_dataset.json').open(encoding='utf-8') as stream:
        first: dict[str, Any] = json.load(stream)
    ds: meta.Dataset = meta.Dataset.from_config(first)
    other: meta.Variable = ds.search_same_dimensions_as(
        ds.variables['simulated_error_karin'])
    assert other.dimensions == ds.variables['simulated_error_karin'].dimensions

    other = meta.Variable.from_config(other.get_config())
    other.dimensions = other.dimensions + ('dummy', )
    with pytest.raises(ValueError):
        ds.search_same_dimensions_as(other)


def test_pickle() -> None:
    """Test pickling."""
    root: pathlib.Path = pathlib.Path(__file__).parent
    with root.joinpath('first_dataset.json').open(encoding='utf-8') as stream:
        data: dict[str, Any] = json.load(stream)
    ds: meta.Dataset = meta.Dataset.from_config(data)
    other: meta.Dataset = pickle.loads(pickle.dumps(ds))
    assert ds == other


def test_missing_variables() -> None:
    """Test missing_variables."""
    root: pathlib.Path = pathlib.Path(__file__).parent
    with root.joinpath('first_dataset.json').open(encoding='utf-8') as stream:
        data: dict[str, Any] = json.load(stream)
    ds: meta.Dataset = meta.Dataset.from_config(data)
    other: meta.Dataset = pickle.loads(pickle.dumps(ds))

    assert len(ds.missing_variables(other)) == 0

    del other.variables['cross_track_distance']
    del other.variables['cycle_number']

    assert set(ds.missing_variables(other)) == {
        'cross_track_distance', 'cycle_number'
    }

    other.variables['XXX'] = other.variables['longitude']
    other.variables['XXX'].name = 'XXX'
    with pytest.raises(ValueError):
        ds.missing_variables(other)


def test_add_variable() -> None:
    """Test adding a variable."""
    ds = meta.Dataset(('x', 'y'), [])
    ds.add_variable(meta.Variable('a', numpy.float64, dimensions=('x', 'y')))

    with pytest.raises(ValueError):
        ds.add_variable(
            meta.Variable('a', numpy.float64, dimensions=('x', 'y')))

    ds.add_variable(meta.Variable('b', numpy.float64, dimensions=('x', )))
    ds.add_variable(meta.Variable('c', numpy.float64, dimensions=('y', )))

    with pytest.raises(ValueError):
        ds.add_variable(
            meta.Variable('d', numpy.float64, dimensions=('a', 'y')))

    with pytest.raises(ValueError):
        ds.add_variable(
            meta.Variable('e', numpy.float64, dimensions=('a', 'b')))

    with pytest.raises(ValueError):
        ds.add_variable(meta.Variable('f', numpy.float64, dimensions=('a', )))

    ds.add_variable(meta.Variable('g', numpy.float64))


def test_select_variables_by_dims() -> None:
    """Test select_variable_by_dims."""
    ds = meta.Dataset(('a', 'b', 'x', 'y'), [])
    ds.add_variable(meta.Variable('a', numpy.float64, dimensions=('x', 'y')))
    ds.add_variable(meta.Variable('b', numpy.float64, dimensions=('x', )))
    ds.add_variable(meta.Variable('c', numpy.float64, dimensions=('y', )))
    ds.add_variable(meta.Variable('d', numpy.float64, dimensions=('a', 'y')))
    ds.add_variable(meta.Variable('e', numpy.float64, dimensions=('a', 'b')))
    ds.add_variable(meta.Variable('f', numpy.float64, dimensions=('a', )))
    ds.add_variable(meta.Variable('g', numpy.float64))

    assert ds.select_variables_by_dims(('x', 'y')) == {'a', 'b', 'c', 'd'}
    assert ds.select_variables_by_dims(('x', )) == {'a', 'b'}
    assert ds.select_variables_by_dims(('y', )) == {'a', 'c', 'd'}
    assert ds.select_variables_by_dims(('a', 'y')) == {'a', 'c', 'd', 'e', 'f'}
    assert ds.select_variables_by_dims(('a', 'b')) == {'d', 'e', 'f'}
    assert ds.select_variables_by_dims(('a', )) == {'d', 'e', 'f'}
    assert ds.select_variables_by_dims(()) == {'g'}
    assert ds.select_variables_by_dims(('z', )) == set()

    assert ds.select_variables_by_dims(('x', 'y'),
                                       predicate=False) == {'e', 'f', 'g'}
    assert ds.select_variables_by_dims(
        ('x', ), predicate=False) == {'c', 'd', 'e', 'f', 'g'}
    assert ds.select_variables_by_dims(
        ('y', ), predicate=False) == {'b', 'e', 'f', 'g'}
    assert ds.select_variables_by_dims(('a', 'y'),
                                       predicate=False) == {'b', 'g'}
    assert ds.select_variables_by_dims(
        ('a', 'b'), predicate=False) == {'a', 'b', 'c', 'g'}
    assert ds.select_variables_by_dims(
        ('a', ), predicate=False) == {'a', 'b', 'c', 'g'}
    assert ds.select_variables_by_dims(
        (), predicate=False) == {'a', 'b', 'c', 'd', 'e', 'f'}
    assert ds.select_variables_by_dims(
        ('z', ), predicate=False) == {'a', 'b', 'c', 'd', 'e', 'f', 'g'}



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Configuration metadata
======================
"""
from __future__ import annotations

from typing import Any, Iterable, Sequence
import abc

import numcodecs.abc
import numpy
import zarr.codecs
import zarr.meta

from .type_hints import DTypeLike

#: Block size limit used with dask arrays. (128 MiB)
BLOCK_SIZE_LIMIT = 134217728


class Pair(abc.ABC):
    """Handle pair key/value.

    Args:
        name: name of the key.
        value: value of the key.
    """
    __slots__ = ('name', 'value')

    def __init__(self, name: str, value: Any) -> None:
        #: Name of the key.
        self.name: str = name
        #: Value of the key.
        self.value: Any = self._encode(value)

    @staticmethod
    def _encode(value: Any) -> Any:
        """Encode an attribute value as something that can be serialized as
        JSON."""
        if isinstance(value, numpy.ndarray):
            return value.tolist()

        if isinstance(value, numpy.generic):
            return value.item()

        return value

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.name!r}, {self.value})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Pair):
            return False
        return self.get_config() == other.get_config()

    def get_config(self) -> tuple[str, Any]:
        """Get the key/value pair configuration."""
        return self.name, self.value

    @staticmethod
    @abc.abstractmethod
    def from_config(data: tuple[str, Any]) -> Pair:
        """Create a new Pair from the given key/value pair configuration."""


class Dimension(Pair):
    """Handle the metadata of a dataset dimension.

    Args:
        name: name of the dimension.
        value: value of the dimension.
    """

    @staticmethod
    def from_config(data: tuple[str, Any]) -> Dimension:
        """Creates a new instance from its metadata.

        Returns:
            Dimension: a new dimension.
        """
        return Dimension(*data)


class Attribute(Pair):
    """Handle the metadata of a dataset attribute.

    Args:
        name: name of the attribute.
        value: value of the attribute.
    """

    @staticmethod
    def from_config(data: tuple[str, Any]) -> Attribute:
        """Create a new instance from its metadata.

        Args:
            data: attribute configuration.

        Returns:
            Attribute: a new attribute.
        """
        return Attribute(*data)


class Variable:
    """Handle the metadata of a dataset variable.

    Args:
        name: Name of the variable.
        dtype: Data type of the variable.
        dimensions: Names of the dimensions of the variable. Defaults to None.
        attrs: Attributes of the variable. Defaults to None.
        compressor: Compression codec for the variable. Defaults to None.
        fill_value: Fill value for the variable. Defaults to None.
        filters: Filters for the variable. Defaults to None.

    Warning:
        If the variable uses filters, the ``fill_value`` parameter must be the
        value that results from decoding the filter. For example, if the filter
        is ``FixedScaleOffset(0, 1000)`` and the desired ``fill_value`` is
        ``65536``, then the ``fill_value`` parameter must be ``65536 / 1000 =
        65.536``.
    """
    __slots__ = ('attrs', 'compressor', 'dimensions', 'dtype', 'fill_value',
                 'filters', 'name')

    def __init__(self,
                 name: str,
                 dtype: DTypeLike,
                 *,
                 dimensions: Sequence[str] | None = None,
                 attrs: Sequence[Attribute] | None = None,
                 compressor: numcodecs.abc.Codec | None = None,
                 fill_value: Any | None = None,
                 filters: Sequence[numcodecs.abc.Codec] | None = None) -> None:
        attrs = attrs or tuple()

        #: Attributes of the variable.
        self.attrs = tuple(attrs)
        #: Compression codec for the variable.
        self.compressor: numcodecs.abc.Codec | None = compressor
        #: Dimensions of the variable.
        self.dimensions = tuple(dimensions or ())
        #: Data type of the variable.
        self.dtype = numpy.dtype(dtype)
        #: Fill value for the variable.
        self.fill_value: Any | None = fill_value
        #: Filter codecs for the variable.
        self.filters = tuple(filters or ())
        #: Variable name.
        self.name: str = name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.name!r})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Variable):
            return False
        return self.get_config() == other.get_config()

    def get_config(self) -> dict[str, Any]:
        """Get the variable metadata.

        Returns:
            variable configuration.
        """
        compressor: numcodecs.abc.Codec | None
        compressor_config: dict[str, None] | None

        compressor = self.compressor
        compressor_config = compressor.get_config(
        ) if compressor is not None else None

        return {
            'attrs': sorted(item.get_config() for item in self.attrs),
            'compressor': compressor_config,
            'dimensions': self.dimensions,
            'dtype': zarr.meta.encode_dtype(self.dtype),
            'fill_value': zarr.meta.encode_fill_value(self.fill_value,
                                                      self.dtype),
            'filters': tuple(item.get_config() for item in self.filters),
            'name': self.name,
        }

    @staticmethod
    def from_config(data: dict[str, Any]) -> Variable:
        """Create a new variable from the given variable configuration.

        Args:
            data: variable configuration.

        Returns:
            new variable.
        """

        def get_codec(codec) -> numcodecs.abc.Codec | None:
            """Get the codec from its configuration."""
            return zarr.codecs.get_codec(codec) if codec is not None else None

        dtype: DTypeLike = zarr.meta.decode_dtype(data['dtype'])
        filters: Sequence[numcodecs.abc.Codec] = tuple(
            zarr.codecs.get_codec(item) for item in data['filters']
            if item is not None)

        return Variable(
            data['name'],
            dtype,
            dimensions=data['dimensions'],
            attrs=tuple(Attribute.from_config(item) for item in data['attrs']),
            compressor=get_codec(data['compressor']),
            fill_value=zarr.meta.decode_fill_value(data['fill_value'], dtype),
            filters=filters,
        )

    def set_for_insertion(self) -> Variable:
        """Create a new variable without any attribute.

        Returns:
            The variable.
        """
        return Variable(self.name,
                        self.dtype,
                        dimensions=self.dimensions,
                        compressor=self.compressor,
                        fill_value=self.fill_value,
                        filters=self.filters)


class Dataset:
    """Handle the metadata of a dataset.

    Args:
        dimensions: A sequence of strings representing the dimensions of the
            dataset.
        variables: A sequence of :py:class:`Variable` objects representing the
            variables of the dataset.
        attrs: An optional sequence of :py:class:`Attribute` objects
            representing the attributes of the dataset. Defaults to None.
        chunks: An optional sequence of :py:class:`Dimension` objects
            representing the chunk size for each dimension. Defaults to None.
        block_size_limit: An optional integer representing the maximum size
            (in bytes) of a block/chunk of variable's data.
    """
    __slots__ = ('dimensions', 'variables', 'attrs', 'chunks',
                 'block_size_limit')

    def __init__(self,
                 dimensions: Sequence[str],
                 variables: Sequence[Variable],
                 *,
                 attrs: Sequence[Attribute] | None = None,
                 chunks: Sequence[Dimension] | None = None,
                 block_size_limit: int | None = None) -> None:
        #: Dimensions of the dataset.
        self.dimensions = tuple(dimensions)

        #: Variables of the dataset.
        self.variables: dict[str, Variable] = {
            item.name: item
            for item in variables
        }

        #: Attributes of the dataset.
        self.attrs = list(attrs or [])

        #: Maximum data chunk size
        self.block_size_limit: int = block_size_limit or BLOCK_SIZE_LIMIT

        #: Chunk size for each dimension
        self.chunks = list(chunks or [])

    def select_variables(
        self,
        keep_variables: Iterable[str] | None = None,
        drop_variables: Iterable[str] | None = None,
    ) -> set[str]:
        """Select variables to keep or drop from the dataset.

        Args:
            keep_variables: A list of variables to retain from the Dataset.
                If None, all variables are kept.
            drop_variables: A list of variables to exclude from the Dataset.
                If None, no variables are dropped.

        Returns:
            The selected variables.
        """
        result = set(self.variables)
        if keep_variables is not None:
            result &= set(keep_variables)
        if drop_variables is not None:
            result -= set(drop_variables)
        return result

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Dataset):
            return False
        return self.get_config() == other.get_config()

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def get_config(self) -> dict[str, Any]:
        """Get the dataset metadata.

        Returns:
            Dataset configuration.
        """
        attrs: list[tuple[str, Any]]
        variables: tuple[dict[str, Any], ...]

        attrs = sorted(item.get_config() for item in self.attrs)
        variables = tuple(self.variables[name].get_config()
                          for name in sorted(self.variables))

        return {
            'attrs': attrs,
            'dimensions': self.dimensions,
            'variables': variables,
            'chunks': tuple(item.get_config() for item in self.chunks),
            'block_size_limit': self.block_size_limit
        }

    def add_variable(self, variable: Variable) -> None:
        """Add a variable to the dataset.

        Args:
            variable: variable to add.

        Raises:
            TypeError: If the variable is not a Variable object.
            ValueError: If the variable already exists in the dataset or if
                the variable's dimensions do not match the dataset's
                dimensions.
        """
        if not isinstance(variable, Variable):
            raise TypeError(
                f'variable must be a Variable, not {type(variable)}')
        if variable.name in self.variables:
            raise ValueError(
                f'The variable {variable.name!r} already exists in the '
                'collection.')
        dimensions = set(self.dimensions)
        # Looking for unknown dimensions.
        if (set(variable.dimensions) | dimensions) != dimensions:
            raise ValueError(
                'The new variable must use the dataset dimensions.')
        self.variables[variable.name] = variable

    @staticmethod
    def from_config(data: dict[str, Any]) -> Dataset:
        """Create a new dataset from the given dataset configuration.

        Args:
            data: dataset configuration.

        Returns:
            New dataset.
        """
        return Dataset(
            dimensions=data['dimensions'],
            variables=tuple(
                Variable.from_config(item) for item in data['variables']),
            attrs=tuple(Attribute.from_config(item) for item in data['attrs']),
            chunks=tuple(
                Dimension.from_config(item)
                for item in data.get('chunks', [])),
            block_size_limit=data.get('block_size_limit'),
        )

    def search_same_dimensions_as(self, variable: Variable) -> Variable:
        """Searches for a variable in this dataset that has the same dimensions
        as the given variable.

        Args:
            variable: The variable used for searching.

        Returns:
            The variable that has the same dimensions as the supplied
            variable.

        Raises:
            ValueError: If no variable with the same dimensions as the given
            variable is found.
        """
        for item in self.variables.values():
            if item.dimensions == variable.dimensions:
                return item
        raise ValueError('No variable using the same dimensions exists.')

    def missing_variables(self, other: Dataset) -> tuple[str, ...]:
        """Finds the variables in the provided dataset that are not in this
        instance.

        Args:
            other: The dataset to compare against.

        Returns:
            A tuple containing the names of the variables that are defined in
            this dataset but not in the provided dataset.

        Raises:
            ValueError: If the provided dataset does not define one or more
                variables that are defined in this dataset.
        """
        this = set(self.variables)
        others = set(other.variables)

        if len(others - this):
            raise ValueError('The reference dataset does not define the '
                             f'{", ".join(others - this)} variables that are '
                             'defined in this dataset.')

        return tuple(this - others)

    def select_variables_by_dims(self,
                                 dims: Sequence[str],
                                 predicate: bool = True) -> set[str]:
        """Select variables that have at least one dimension in the given
        dimensions depending on the predicate.

        Args:
            dims: A sequence of dimensions to select.
            predicate: A boolean value that determines whether to select
                variables that have the given dimensions (True) or variables
                that don't have the given dimensions (False).

        Returns:
            A set of variable names that have the given dimensions (if predicate
            is True) or don't have the given dimensions (if predicate is False).
        """
        if len(dims) == 0:
            return {
                name
                for name, var in self.variables.items()
                if (len(var.dimensions) == 0) == predicate
            }

        set_of_dims = set(dims)
        return {
            name
            for name, var in self.variables.items()
            if bool(set(var.dimensions) & set_of_dims) == predicate
        }



"""
Overview of a Collection.
=========================

This section outlines the steps required to get started with the main features
of a ``Collection``.
"""
from __future__ import annotations

from typing import Iterator
import datetime
import pprint

import dask.distributed
import fsspec
import numpy

import zcollection
import zcollection.tests.data


# %%
# Initialization of the environment
# ---------------------------------
#
# Before we create our first collection, we will create a dataset to record.
def create_dataset() -> zcollection.Dataset:
    """Create a dataset to record."""
    generator: Iterator[zcollection.Dataset] = \
        zcollection.tests.data.create_test_dataset_with_fillvalue()
    return next(generator)


zds: zcollection.Dataset | None = create_dataset()
assert zds is not None
zds.to_xarray()

# %%
# We will create the file system that we will use. In this example, a file
# system in memory.
fs: fsspec.AbstractFileSystem = fsspec.filesystem('memory')

# %%
# Finally we create a local dask cluster using only threads in order to work
# with the file system stored in memory.
cluster = dask.distributed.LocalCluster(processes=False)
client = dask.distributed.Client(cluster)

# %%
# Creation of the partitioning
# ----------------------------
#
# Before creating our collection, we define the partitioning of our dataset. In
# this example, we will partition the data by ``month`` using the variable
# ``time``.
partition_handler = zcollection.partitioning.Date(('time', ), resolution='M')

# %%
# Finally, we create our collection:
collection: zcollection.Collection = zcollection.create_collection(
    'time', zds, partition_handler, '/my_collection', filesystem=fs)

# %%
# .. note::
#
#    The collection created can be accessed using the following command ::
#
#        >>> collection = zcollection.open_collection("/my_collection",
#        >>>                                          filesystem=fs)
#
# When the collection has been created, a configuration file is created. This
# file contains all the metadata to ensure that all future inserted data will
# have the same features as the existing data (data consistency).
pprint.pprint(collection.metadata.get_config())

# %%
# Now that the collection has been created, we can insert new records.
collection.insert(zds)

# %%
# .. note::
#
#     When inserting it's possible to specify the :ref:`merge strategy of a
#     partition <merging_datasets>`. By default, the last inserted data
#     overwrite the existing ones. Others strategy can be defined, for example,
#     to update existing data (overwrite the updated data, while keeping the
#     existing ones). This last strategy allows updating incrementally an
#     existing partition. ::
#
#         >>> import zcollection.merging
#         >>> collection.insert(
#         ...     ds, merge_callable=zcollection.merging.merge_time_series)
#
# Let's look at the different partitions thus created.
pprint.pprint(fs.listdir('/my_collection/year=2000'))

# %%
# This collection is composed of several partitions, but it is always handled
# as a single data set.
#
# Loading data
# ------------
# To load the dataset call the method
# :py:meth:`load<zcollection.collection.Collection.load>` on the instance.  By
# default, the method loads all partitions stored in the collection.
collection.load(delayed=True)

# %%
# .. note::
#
#   By default, the data is loaded as a :py:class:`dask.array<da.Array>`. It is
#   possible to load the data as a :py:class:`numpy.ndarray` by specifying the
#   parameter ``delayed=False``.
#
# You can also filter the partitions to be considered by filtering the
# partitions using keywords used for partitioning in a valid Python expression.
collection.load(filters='year == 2000 and month == 2')

# %%
# You can also used a callback function to filter partitions with a complex
# condition.
collection.load(
    filters=lambda keys: datetime.date(2000, 2, 15) <= datetime.date(
        keys['year'], keys['month'], 1) <= datetime.date(2000, 3, 15))

# %%
# Note that the :py:meth:`load<zcollection.collection.Collection.load>`
# function may return None if no partition has been selected.
assert collection.load(filters='year == 2002 and month == 2') is None

# %%
# Editing variables
# -----------------
#
# .. note::
#
#     The functions for modifying collections are not usable if the collection
#     is :py:meth:`open<zcollection.open_collection>` in read-only mode.
#
# It's possible to delete a variable from a collection.
collection.drop_variable('var2')
collection.load()

# %%
# The variable used for partitioning cannot be deleted.
try:
    collection.drop_variable('time')
except ValueError as exc:
    print(exc)

# %%
# The :py:meth:`add_variable<zcollection.collection.Collection.add_variable>`
# method allows you to add a new variable to the collection.
collection.add_variable(zds.metadata().variables['var2'])

# %%
# The newly created variable is initialized with its default value.
zds = collection.load()
assert zds is not None
zds.variables['var2'].values


# %%
# Finally it's possible to
# :py:meth:`update<zcollection.collection.Collection.update>` the existing
# variables.
#
# In this example, we will alter the variable ``var2`` by setting it to 1
# anywhere the variable ``var1`` is defined.
def ones(zds) -> dict[str, numpy.ndarray]:
    """Returns a variable with ones everywhere."""
    return dict(var2=zds.variables['var1'].values * 0 + 1)


collection.update(ones)  # type: ignore[arg-type]

zds = collection.load()
assert zds is not None
zds.variables['var2'].values


# %%
# ..note::
#
#   The method :py:meth:`update<zcollection.collection.Collection.update>`
#   supports the ``delayed`` parameter. If ``delayed=True``, the function
#   ``ones`` is applied to each partition using a Dask array as container
#   for the variables data stored in the provided dataset. This is the default
#   behavior. If ``delayed=False``, the function ``ones`` is applied to each
#   partition using a Numpy array as container.
#
# Sometime is it important to know the values of the neighboring partitions.
# This can be done using the
# :py:meth:`update<zcollection.collection.Collection.update>` method with the
# ``depth`` argument. In this example, we will set the variable ``var2`` to 2
# everywhere the processed partition is surrounded by at least one partition, -1
# if the left partition is missing and -2 if the right partition is missing.
#
# .. note::
#
#   ``partition_info`` contains information about the target partition: a tuple
#   with the partitioned dimension and the slice to select the partition. If the
#   start of the slice is 0, it means that the left partition is missing. If the
#   stop of the slice is equal to the length of the given dataset, it means that
#   the right partition is missing.
def twos(ds, partition_info: tuple[str, slice]) -> dict[str, numpy.ndarray]:
    """Returns a variable with twos everywhere if the partition is surrounded
    by partitions on both sides, -1 if the left partition is missing and -2 if
    the right partition is missing."""
    data = numpy.zeros(ds.variables['var1'].shape, dtype='int8')
    dim, indices = partition_info
    assert dim == 'num_lines'
    if indices.start != 0:
        data[:] = -1
    elif indices.stop != data.shape[0]:
        data[:] = -2
    else:
        data[:] = 2
    return dict(var2=data)


collection.update(twos, depth=1)  # type: ignore[arg-type]

zds = collection.load()
assert zds is not None
zds.variables['var2'].values

# %%
# Map a function over the collection
# ----------------------------------
# It's possible to map a function over the partitions of the collection.
for partition, array in collection.map(lambda ds: (  # type: ignore[arg-type]
        ds['var1'].values + ds['var2'].values)).compute():
    print(f' * partition = {partition}: mean = {array.mean()}')

# %%
# .. note::
#
#     The :py:meth:`map<zcollection.collection.Collection.map>` method is
#     lazy. To compute the result, you need to call the method ``compute``
#     on the returned object.
#
# It's also possible to map a function over the partitions with a a number of
# neighboring partitions, like the
# :py:meth:`update<zcollection.collection.Collection.update>`. To do so, use the
# :py:meth:`map_overlap<zcollection.collection.Collection.map_overlap>` method.
for partition, array in collection.map_overlap(
        lambda ds: (  # type: ignore[arg-type]
            ds['var1'].values + ds['var2'].values),
        depth=1).compute():
    print(f' * partition = {partition}: mean = {array.mean()}')

# %%
# Close the local cluster to avoid printing warning messages in the other
# examples.
client.close()
cluster.close()



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""This script is the entry point for building, distributing and installing
this module using distutils/setuptools."""
import pathlib

import setuptools
import setuptools.command.sdist

# Working directory
WORKING_DIRECTORY = pathlib.Path(__file__).parent.absolute()


class SDist(setuptools.command.sdist.sdist):
    """Custom sdist command that copies the pytest configuration file into the
    package."""
    user_options = setuptools.command.sdist.sdist.user_options

    def run(self):
        """Carry out the action."""
        source = WORKING_DIRECTORY.joinpath('conftest.py')
        target = WORKING_DIRECTORY.joinpath('zcollection', 'conftest.py')
        source.rename(target)
        try:
            super().run()
        finally:
            target.rename(source)


setuptools.setup(cmdclass={'sdist': SDist})



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
==========
Test setup
==========
"""


def pytest_addoption(parser) -> None:
    """Add command line options to pytest."""
    parser.addoption(
        '--s3',
        action='store_true',
        default=False,
        help='Enable tests on the local S3 server driven by minio. '
        '(default: False)')
    parser.addoption(
        '--memory',
        action='store_true',
        default=False,
        help='Use a file system in memory instead of the local file system. '
        '(default: False)')
    parser.addoption(
        '--threads_per_worker',
        action='store',
        default=None,
        type=int,
        help='Number of threads for each worker Dask. (default: the number of '
        'logical cores of the target platform).')
    parser.addoption(
        '--n_workers',
        action='store',
        default=None,
        type=int,
        help='Number of core for each worker Dask. (default: the number of '
        'cores of the target platform).')
    parser.addoption(
        '--processes',
        action='store_true',
        default=False,
        help='Whether to use processes or threads for Dask. (default: False)')



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Create test variables.
======================
"""
import numpy
import zarr

from .. import Array, DelayedArray
from ... import meta


def array(name='var1', fill_value=0) -> Array:
    """Creates a test variable with the given name, fill value, dimensions, and
    attributes.

    Args:
        name: The name of the variable.
        fill_value: The fill value for uninitialized parts of the array.

    Returns:
        An Array object.
    """
    return Array(name=name,
                 data=numpy.arange(10, dtype='int64').reshape(5, 2),
                 dimensions=('x', 'y'),
                 attrs=(meta.Attribute(name='attr', value=1), ),
                 compressor=zarr.Blosc(cname='zstd', clevel=1),
                 fill_value=fill_value,
                 filters=(zarr.Delta('int64',
                                     'int32'), zarr.Delta('int32', 'int32')))


def delayed_array(name='var1', fill_value=0) -> DelayedArray:
    """Create a delayed test variable with the given name, fill value,
    dimensions, and attributes.

    Args:
        name: The name of the variable.
        fill_value: The fill value for uninitialized parts of the array.

    Returns:
        A DelayedArray object representing a lazily-evaluated test variable.
    """
    return DelayedArray(name=name,
                        data=numpy.arange(10, dtype='int64').reshape(5, 2),
                        dimensions=('x', 'y'),
                        attrs=(meta.Attribute(name='attr', value=1), ),
                        compressor=zarr.Blosc(cname='zstd', clevel=1),
                        fill_value=fill_value,
                        filters=(zarr.Delta('int64', 'int32'),
                                 zarr.Delta('int32', 'int32')))



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
In memory variable arrays.
==========================
"""
from __future__ import annotations

from typing import Any, Sequence

import dask.array.core
import dask.array.ma
import numcodecs.abc
import numpy
import zarr

from ..meta import Attribute
from ..type_hints import ArrayLike, NDArray, NDMaskedArray
from .abc import Variable, concat, new_variable, not_equal


def _as_numpy_array(
    arr: Any,
    *,
    fill_value: Any | None = None,
) -> tuple[NDArray, Any]:
    """Convert an array-like object to a numpy array.

    Args:
        arr: An array-like object.
        fill_value: The fill value.

    Returns:
        If the data provided is a masked array, the functions return the array
        with masked data replaced by its fill value and the fill value of the
        offered masked array. Otherwise, the provided array and fill value.
    """
    result: NDArray = numpy.asanyarray(arr)
    if isinstance(result, numpy.ma.MaskedArray):
        if fill_value is not None and not_equal(fill_value, result.fill_value):
            raise ValueError(
                f'The fill value {fill_value!r} does not match the fill value '
                f'{result.fill_value!r} of the array.')
        return numpy.ma.filled(result, result.fill_value), result.fill_value
    return result, fill_value


class Array(Variable):
    """Access to the chunked data using numpy arrays.

    Args:
        name: Name of the variable
        data: Variable data
        dimensions: Variable dimensions
        attrs: Variable attributes
        compressor: Compression codec
        fill_value: Value to use for uninitialized values
        filters: Filters to apply before writing data to disk
    """

    def __init__(self,
                 name: str,
                 data: ArrayLike[Any],
                 dimensions: Sequence[str],
                 *,
                 attrs: Sequence[Attribute] | None = None,
                 compressor: numcodecs.abc.Codec | None = None,
                 fill_value: Any | None = None,
                 filters: Sequence[numcodecs.abc.Codec] | None = None) -> None:
        array: NDArray
        array, fill_value = _as_numpy_array(data, fill_value=fill_value)
        super().__init__(
            name,
            array,
            dimensions,
            attrs=attrs,
            compressor=compressor,
            fill_value=fill_value,
            filters=filters,
        )

    @property
    def data(self) -> dask.array.core.Array:
        """Return the numpy array wrapped in a dask array. If the variable has
        a fill value, the result is a masked array where masked values are
        equal to the fill value.

        Returns:
            The dask array

        .. seealso::

            :meth:`Variable.array`
        """
        if self.fill_value is None:
            return dask.array.core.from_array(self.array)
        return dask.array.ma.masked_equal(self.array, self.fill_value)

    @property
    def values(self) -> NDArray | NDMaskedArray:
        """Return the variable data as a numpy array.

        .. note::

            If the variable has a fill value, the result is a masked array where
            masked values are equal to the fill value.

        Returns:
            The variable data
        """
        return self.array if self.fill_value is None else numpy.ma.masked_equal(
            self.array, self.fill_value)

    @values.setter
    def values(self, data: Any) -> None:
        """Defines the underlying numpy array. If the data provided is a masked
        array, it's converted to an array, where the masked values are replaced
        by its fill value, and its fill value becomes the new fill value of
        this instance. Otherwise, the underlying array is defined as the new
        data and the fill value is set to None.

        Args:
            data: The new data to use

        Raises:
            ValueError: If the shape of the data does not match the shape of
                the stored data.
        """
        if len(data.shape) != len(self.dimensions):
            raise ValueError('data shape does not match variable dimensions')
        self.array, self.fill_value = _as_numpy_array(
            data, fill_value=self.fill_value)

    def persist(self, **_) -> Array:
        """Persist the variable data into memory.

        Returns:
            The variable
        """
        return self

    def compute(self, **_) -> NDArray | NDMaskedArray:
        """Return the variable data as a numpy array.

        .. note::

            If the variable has a fill value, the result is a masked array where
            masked values are equal to the fill value.
        """
        return self.values

    def fill(self) -> Array:
        """Fill the variable with the fill value. If the variable has no fill
        value, this method does nothing.

        Returns:
            The variable.
        """
        if self.fill_value is not None:
            self.array = numpy.full_like(self.array, self.fill_value)
        return self

    @classmethod
    def from_zarr(cls, array: zarr.Array, name: str, dimension: str,
                  **kwargs) -> Array:
        """Create a new variable from a zarr array.

        Args:
            array: The zarr array
            name: Name of the variable
            dimension: Name of the attribute that defines the dimensions of the
                variable
            **kwargs: Additional arguments. These arguments are ignored, but
                they are accepted to be compatible with the base class.

        Returns:
            The variable
        """
        attrs = tuple(
            Attribute(k, v) for k, v in array.attrs.items() if k != dimension)
        return new_variable(cls,
                            name=name,
                            array=array[...],
                            dimensions=array.attrs[dimension],
                            attrs=attrs,
                            compressor=array.compressor,
                            fill_value=array.fill_value,
                            filters=tuple(array.filters or ()))

    def concat(self, other: Array | Sequence[Array], dim: str) -> Array:
        """Concatenate this variable with another variable or a list of
        variables along a dimension.

        Args:
            other: Variable or list of variables to concatenate with this
                variable.
            dim: Dimension to concatenate along.

        Returns:
            New variable.

        Raises:
            ValueError: if the variables provided is an empty sequence or if
                any item in the sequence is not an instance of :class:`Array`.
        """
        return concat(self, other, numpy.concatenate, dim)

    def __getitem__(self, key: Any) -> Any:
        """Get a slice of the variable.

        Args:
            key: Slice or index to use.
        Returns:
            The variable slice.
        """
        return self.array[key]

    def rechunk(self, **_) -> Array:
        """Rechunk the variable.

        Returns:
            The variable.
        """
        return self
